#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:expandtab:sw=4:tw=0
###############################################################################
# component.py.py - component parts                                           #
#                                                                             #
# $Id: component.py 193 2007-06-06 15:59:54Z smrcka $
#                                                                             #
# Author: Ales Smrcka <smrcka@fit.vutbr.cz>                                   #
#                                                                             #
# This program is free software; you can redistribute it and/or modify        #
#   it under the terms of the GNU General Public License as published by      #
#   the Free Software Foundation; either version 2 of the License, or         #
#   (at your option) any later version.                                       #
#   See http://www.gnu.org/licenses/gpl.html for more details.                #
#                                                                             #
###############################################################################
import simplify

# map<id,GlobalObject> (id is lower string): this is a super id_map,
# furthermore there is id_map inside every VHDLobject
id_map={}

# dependency graph: id->set of id (dependencies - id depends on ids)
#TODO!
dep_graph={}

# ignore list: list of IgnoreEntity (automatically generated by IgnoreEntity)
ignore_list=[]

# CAException (super class for further exceptions within this project)
class CAException(Exception):
    def print_error(self):
        pass

# exception for VHDL syntax errors
class VHDLException(CAException):
    def print_error(self):
        if len(self.args)>=4:
            error,filename,lineno = self.args[1:4]
            f=open(filename)
            line=f.readlines()[lineno-1].strip()
            f.close()
            # GCC error style
            print '%s: At: %s' % (filename, line)
            print '%s:%d: error: %s' % (filename, lineno, error)
        else:    
            print 'error:%s' % self.args[0]

# exception for transformation errors
class TransException(CAException):
    def print_error(self):
        if len(self.args)>=4:
            error,filename,lineno = self.args[1:4]
            print '%s: Transformation error' % filename
            print '%s:%d: error: %s' % (filename, lineno, error)
        else:    
            print 'error:%s' % self.args[0]

# class of "ignore part" of any VHDLobject
class IgnoreEntity(object):
    owner=None      # VHDLobject
    body=None       # body of ignoring part
    comment=''      # comment
    def __init__(self, owner=None, body=None, comment=''):
        self.owner=owner
        self.body=body
        self.comment=comment

    def __str__(self):
        if self.owner:
            return '%s.%s(%s)' % \
                (str(self.owner), type(self).__name__, self.comment)
        else:
            return ''

###############################################################################
# superclass of all VHDL objects
class VHDLobject(object):
    "Superclass of all vhdl part classes"
    # container of all child VHDLobjects
    container=[] # currently unsupported

    # id_map ⊆ str -> IdObject
    id_map={}

    # owner of this object (parent in syntax hierarchy
    # e.g. entity<-architecture<-procedure<-variable
    parent=None

    # for declaration production: the first VHDLobject in syntax stack with
    # declaring==True is the object to declare a new declaration statement
    declaring=False

    # source of this object
    filename=''
    lineno=0

    # set source filename and line number from p (yacc/lex parameter) and
    # any other attribute
    def __init__(self, p=None, **kargs):
        self.container=[]
        self.id_map={}
        if p:
            try:
                self.lineno=p.lineno(1)
                if not self.lineno:
                    self.lineno=p.lexer.lineno
            except:
                self.lineno=p.lexer.lineno
            self.filename=p.lexer.filename
        kargs=self._normalize_kargs(kargs)
        if kargs:
            if 'kargs' in kargs.keys():
                kargs=kargs['kargs']
        if kargs:
            for key,value in kargs.iteritems():
                self.__setattr__(key, value)

    def _normalize_kargs(self, kargs):
        if kargs:
            while 'kargs' in kargs.keys():
                kargs=kargs['kargs']
        return kargs

    # vhdlobject[str_id] returns IdObject (or None if not found) which is
    # declared within this namespace - declared here or in some parent
    # object
    def __getitem__(self, key):
        """search for id key in current and parent entity namespace, returns
        VHDLobject or None"""
        if self.id_map.has_key(key):
            return self.id_map[key]
        elif VHDLobject in type(self.parent).mro():
            return self.parent[key]
        else:
            return None

    # take care of parent attribute
    def __setattr__(self, key, value):
        super(VHDLobject, self).__setattr__(key, value)
        if key=='parent':
            if VHDLobject not in type(value).mro():
                self.vhdl_error('confused with parent %s' % value)
            elif IdObject in type(self).mro():
                self.parent.register(self)

    # returns string representation of this object
    # filename:lineno:type of an object or only the type
    def __str__(self):
        where=self.where()
        if where:
            return '%s:%s' % (self.where(), type(self).__name__)
        else:
            return type(self).__name__
    
    def find(self, path=''):
        ppath=path.split('/')
        if ppath[0]:
            name=ppath[0]
            path='/'.join(ppath[1:])
            if name.lower()!=name:
                # looking for type 'name'
                for o in self.id_map.itervalues():
                    if name==type(o).__name__:
                        return o.find(path)
                return None
            else:
                # looking for object with id 'name'
                try:
                    return self.id_map[name].find(path)
                except:
                    return None
        else:
            return self

    def ignore(self, body=None, comment='unsupported'):
        "ignore(body, comment) adds part of VHDLobject to ignore_list"
        ignore_list.append(IgnoreEntity(self, body, comment))
    
    def ignore_parts(self):
        "ignore_parts() returns ignoring entyties of this object"
        return [i for i in ignore_list if i.owner==self]

    def register(self, idobject):
        """register(idobject) registers idobject.id to id_map,
           idobject could be a list of idobjects"""
        if IdObject in type(idobject).mro():
            # if there exists another IdObject with the same ID...
            if self.id_map.has_key(idobject.id) and \
               self.id_map[idobject.id] is not idobject:
                # raise an error
                idobject.vhdl_error( \
                    'unable to register %s, already registered as %s' % \
                    (idobject.id, self.id_map[idobject.id]))
            else: # ...else register a new IdObject
                self.id_map[idobject.id]=idobject
        elif type(idobject)==list:
            # register a list of IdObjects
            for i in idobject:
                self.register(i)
        else:
            err_msg = 'unable to register object, type %s unsupported' % \
                type(idobject)
            if VHDLobject in type(idobject).mro():
                idobject.vhdl_error(err_msg)
            else:
                self.vhdl_error(err_msg)

    def supported(self):
        "supported() checks whether a VHDLobject is not marked as unsupported"
        return self.__doc__ != "unsupported"
        
    def trans_error(self, s):
        where=self.where()
        if where:
            raise TransException('%s:%s' % (where, s), \
                                 s, self.filename, self.lineno)
        else:
            raise TransException(s, s, self.filename, self.lineno)
            

    def unsupport(self, part=None):
        "unsupport() set the doc part of object to 'unsupported'"
        self.__doc__='unsupported'
        self.unsupported_part=part

    def vhdl_error(self, s):
        # TODOv2: enlist the exception, do not raise it immediately
        where=self.where()
        if where:
            raise VHDLException('%s:%s' % (where, s), \
                                s, self.filename, self.lineno)
        else:
            raise VHDLException(s, s, self.filename, self.lineno)

    def where(self):
        "where() returns string `filename:lineno' of object source"
        if self.filename:
            return '%s:%d' % (self.filename, self.lineno)
        else:
            return ''
    
###############################################################################
# Syntax special classes (most probably abstract blocks)

class IdObject(VHDLobject):
    id=''
    def __str__(self):
        return '%s(%s)' % (type(self).__name__, self.id)
    
class GlobalObject(IdObject):
    def __setattr__(self, key, value):
        # take care of 'id' attribute
        global id_map
        if key=='id':
            if id_map.has_key(value) and id_map[value] is not self:
                self.vhdl_error( \
                    'unable to register %s, already registered as %s' % \
                    (value, id_map[value]))
            else:
                id_map[value]=self
        super(GlobalObject, self).__setattr__(key, value)

class EvalObject(VHDLobject):
    # list of Usage
    usage=None

    def __init__(self, p=None, **kargs):
        self.usage=[]
        super(EvalObject, self).__init__(p, kargs=kargs)

    def eval(self, prev_values, values):
        """eval(prev_values, values) evaluate expression (this is valid
        only on concrete class, not abstract Expr"""
        raise VHDLExpception( \
            '[internal] evaluation can not be performed on %s object' % \
            type(self).__name__)

    def get_type(self):
        raise VHDLException( \
            '[internal] get_type can not be called on %s object' % \
            type(self).__name__)

class CommonType(VHDLobject):
    def codomain(self):
        return []
    def codomain_count(self):
        return 0

NoType=CommonType()

class PredefinedType(CommonType):
    pass

##############################################################################
# common objects
class Parameter(IdObject, EvalObject):
    type=''
    value=None
    constraint=None

    def __str__(self):
        if self.value:
            return 'Parameter(%s %s = %s)' % (self.type, self.id, self.value)
        else:
            return 'Parameter(%s %s)' % (self.type, self.id)

    def eval(self, prev_values, values):
        """eval(prev_values, values) return value of parameter (if
        defined in values or if has set a value)"""
        if values.has_key(self.id):
            return values[self.id]
        elif self.value != None:
            return int(self.value)
        else:
            return self

    def get_type(self):
        # TODO supported types
        if self.type=='integer':
            return Int
        else:
            return NoType

class Time(VHDLobject):
    value=0     # value of time
    unit=''     # one of: 'hr','min','sec','ms','us','ns','ps','fs'
    def __str__(self):
        return 'Time(%s %s)' % (self.value, self.unit)

class Range(VHDLobject):
    bottom=None # Expr
    top=None    # Expr
    def __str__(self):
        return 'Range(%s, %s)' % (self.bottom, self.top)

class BoolType(PredefinedType):
    def codomain(self):
        return [0,1]

    def codomain_count(self):
        return 2

Bool=BoolType()

class BinVectorType(PredefinedType):
    "unsupported"
    def __str__(self):
        try:
            return '%s[%s:%s]' % (type(self).__name__, \
                self.range.bottom, self.range.top)
        except:
            return type(self).__name__

class IntType(PredefinedType):
    pass

Int=IntType()

##############################################################################
# library units
class Library(GlobalObject):
    packages=[]     # list of Package

    def __init__(self, p=None, **kargs):
        self.packages=[]
        super(Library, self).__init__(p, kargs=kargs)

class Package(VHDLobject):
    use=''

class PackageUnit(GlobalObject):
    "unsupported"

class PackageBody(GlobalObject):
    "unsupported"

class Configuration(GlobalObject):
    "unsupported"
    entity=''       # configuration of entity

class BlockConfig(VHDLobject):
    "unsupported"
    label=''

class ConfConfig(VHDLobject):
    "unsupported"

class Entity(GlobalObject):
    parameters=[]   # list of Parameter
    ports=[]        # list of Port
    architecture=None   # body of entity

    def __init__(self, p=None, **kargs):
        self.parameters=[]
        self.ports=[]
        super(Entity, self).__init__(p, kargs=kargs)

    def find(self, path=''):
        r=super(Entity, self).find(path)
        if r==None:
            return super(Entity, self).find('Architecture/%s' % path)
        else:
            return r

    def register(self, idobject):
        super(Entity, self).register(idobject)
        if Parameter in type(idobject).mro():
            self.parameters.append(idobject)
        elif Port in type(idobject).mro():
            self.ports.append(idobject)

class Architecture(GlobalObject):
    # architecture declaration
    signals=[]      # list of Signal
    constants=[]    # list of Constant
    # architecture body
    connections=[]  # list of CondAssign
    processes=[]    # list of Process

    def __init__(self, p=None, **kargs):
        self.signals=[]
        self.constants=[]
        self.connections=[]
        self.processes=[]
        super(Architecture, self).__init__(p, kargs=kargs)

    def register(self, idobject):
        super(Architecture, self).register(idobject)
        if Signal in type(idobject).mro():
            self.signals.append(idobject)
        elif Constant in type(idobject).mro():
            self.constants.append(idobject)

##############################################################################
# declaration

class Signal(IdObject, EvalObject):
    # type either Expr or one of CommonType
    type=None
    value=None      # Expr: value optionally assigned

    # flag:
    # '' = nothing special
    # 'clock' = there is a monitor of edge of this signal
    flag=''

    # sensitivity list (actually set of Signal)
    senslist=None

    def __init__(self, p=None, **kargs):
        self.senslist=[]
        super(Signal, self).__init__(p, kargs=kargs)

    def __str__(self):
        return 'Signal(%s %s)' % (self.type, self.id)
    
    def eval(self, prev_values, values):
        """eval(prev_values, values) return value of signal (values and
        prev_values set values of Signals)"""
        if self.id in values.keys():
            return values[self.id]
        elif self.value:
            return self.value.eval(prev_values, values)
        else:
            return self # nothing known about this signal

    def get_type(self):
        if self.type:
            if CommonType in type(self.type).mro():
                return self.type
            else:
                self.vhdl_error("type '%s' not validated" % self.type)
                return NoType
        else:
            self.vhdl_error('type not specified')
            return NoType

class Port(Signal):
    direction='inout'   # one of 'in', 'out', 'inout'
    def __str__(self):
        return 'Port(%s %s)' % (self.type, self.id)

class SignalEvent(EvalObject):
    signal=None    # Signal'event representative
    def __str__(self):
        return "Event(%s)" % (self.signal.id)
    
    # take care of 'signal' attribute, set the 'clock' flag of it
    def __setattr__(self, key, value):
        super(SignalEvent, self).__setattr__(key, value)
        if key=='signal':
            self.signal.flag='clock'

    def eval(self, prev_values, values):
        """eval(prev_values, values) return value of signal (values and
        prev_values set values of Signals)"""
        if (not self.signal) or (Signal not in type(self.signal).mro()):
            return None
        id=self.signal.id
        if id in values.keys():
            if id in prev_values.keys():
                if values[id]!=prev_values[id]:
                    return 1
                else:
                    return 0
            else:
                return 0
        else:
            return self # nothing known about this signal

    def get_type(self):
        return Bool

class SignalPosedge(SignalEvent):
    def __str__(self):
        return "Posedge(%s)" % (self.signal.id)

class SignalNegedge(SignalEvent):
    def __str__(self):
        return "Negedge(%s)" % (self.signal.id)

class Constant(IdObject, EvalObject):
    # type either Expr or one of CommonType
    type=None
    value=None      # Expr: value of constant
    def __str__(self):
        if self.value:
            return 'Constant(%s %s = %s)' % (self.type, self.id, self.value)
        else:
            return 'Constant(%s %s)' % (self.type, self.id)
    
    def eval(self, prev_values, values):
        """eval(prev_values, values) return value of signal (values and
        prev_values set values of Signals)"""
        if self.id in values.keys():
            return values[self.id]
        elif self.value:
            return self.value.eval(prev_values, values)
        else:
            return self # nothing known about this signal
    
    def get_type(self):
        if self.type:
            if CommonType in type(self.type).mro():
                return self.type
            else:
                self.vhdl_error("type '%s' not validated" % self.type)
                return NoType
        else:
            self.vhdl_error('type not specified')
            return NoType

class Usage(VHDLobject):
    "an usage of signal or constant"
    access=''       # either Usage.R or Usage.W
    R='read'        # constant for read access
    W='write'       # constant for write access

    def __str__(self):
        return '%s: %s access' % (self.where(), self.access)

class Variable(IdObject, EvalObject):
    "unsupported"

class Type(IdObject):
    "unsupported"

class SubType(IdObject):
    "unsupported"

class File(IdObject):
    "unsupported"

class Alias(IdObject):
    "unsupported"

class Attribute(IdObject):
    "unsupported"

class Component(IdObject):
    "unsupported"

class Function(IdObject):
    "unsupported"

class Procedure(IdObject):
    "unsupported"

class UseEntity(VHDLobject):
    "unsupported"

class UseConfiguration(VHDLobject):
    "unsupported"

##############################################################################
# statements
class Connection(VHDLobject):
    signal=None     # signal (Signal object) to be defined
    expr=None       # circuit to connect to signal (Expr)

class Process(VHDLobject):
    sens_list=[]    # sensitivity list (list of Signal)
    assigns=[]      # list of CondAssign
    
    def __init__(self, p=None, **kargs):
        self.sens_list=[]
        self.assigns=[]
        super(Process, self).__init__(p, kargs=kargs)

class CondConnection(VHDLobject):
    "unsupported"

class CondGenerate(VHDLobject):
    "unsupported"

class Block(VHDLobject):
    "unsupported"

class ProcCall(VHDLobject):
    "unsupported"

class ComponentInstance(VHDLobject):
    "unsupported"

class EntityImpl(VHDLobject):
    "unsupported"

class ForGenerate(VHDLobject):
    "unsupported"

class VarAssign(VHDLobject):
    "unsupported"

class SeqProcCall(VHDLobject):
    "unsupported"

class Loop(VHDLobject):
    "unsupported"

class For(VHDLobject):
    "unsupported"

class Wait(VHDLobject):
    "unsupported"

class Assert(VHDLobject):
    "unsupported"

class Report(VHDLobject):
    "unsupported"

class Next(VHDLobject):
    "unsupported"

class Exit(VHDLobject):
    "unsupported"

class Return(VHDLobject):
    "unsupported"

class Assign(VHDLobject):
    # 'signal' attribute must be assign AFTER initialization; validate() is
    # called during 'signal' assignment, which requires 'parent' attribute to
    # be set.
    signal=None     # Signal to assign a new value states: str, Signal (or Port)
    value=None      # Expr

    def __str__(self):
        return '%s:=%s' % (self.signal.id, self.value)

    # take care of 'signal' attribute, automatically search for Signal object
    def __setattr__(self, key, value):
        super(Assign, self).__setattr__(key, value)
        if key=='signal':
            self.validate()

    def validate(self):
        "validate() validates signal type (str id->VHDLobject)"
        if type(self.signal)==str:
            if not self.parent:
                self.vhdl_error('parent not specified')
            elif VHDLobject not in type(self.parent).mro():
                self.vhdl_error('parent type %s unsupported' % \
                    type(self.parent))
            elif Signal not in type(self.parent[self.signal]).mro():
                self.vhdl_error('signal expected instead of %s' % \
                    self.signal.id)
            else:
                self.signal=self.parent[self.signal]
        elif VHDLobject not in type(self.signal).mro():
            self.vhdl_error('such an assignment not supported: %s' % \
                (self.signal))
            
class CondAssign(VHDLobject):
    guard=None      # Expr of current state guard
    action=None     # Assign
    sens_list=None

    def __init__(self, p=None, **kargs):
        self.sens_list=[]
        super(CondAssign, self).__init__(p, kargs=kargs)

    def __str__(self):
        return '%s -> %s' % (self.guard, self.action)

###############################################################################
# expression


operator_map={
    'and':'and',
    '&':'and',
    '&&':'and',
    'or':'or',
    '|':'or',
    '||':'or',
    'not':'not',
    '!':'not',
    '~':'not',
    '=':'==',
    '==':'==',
    '>=':'>=',
    '<=':'<=',
    '>':'>',
    '<':'<',
    '!=':'!=',
    '/=':'!=',
    '+':'+',
    '-':'-',
    '*':'*',
    'abs':'abs',
    'div':'div',
    '/':'div',
    'mod':'mod',
    'rem':'mod',
    }

# Expr is the superclass for OpExpr and Prim.
class Expr(EvalObject):
    def get_list(self):
        "returns a list from expression"
        pass

    def get_evals(self):
        return simplify.get_used_objects(self.get_list())

class OpExpr(Expr):
    operator=None   # one of operators
    operands=[]     # list of Expr

    def __init__(self, p=None, **kargs):
        self.operands=[]
        super(OpExpr, self).__init__(p, kargs=kargs)

    def __setattr__(self, key, value):
        if key=='operator':
            if value in operator_map.keys():
                self.__dict__[key]=operator_map[value]
            else:
                self.trans_error('unsupported operator %s' % value)
        else:
            super(OpExpr, self).__setattr__(key, value)

    def __str__(self):
        operands=[str(i) for i in self.operands]
        return '(%s,%s)' % (self.operator, ','.join(operands))

    # returns list [operator, operand1, operand2, ...]
    def get_list(self):
        return [self.operator]+[o.get_list() for o in self.operands]

# Prim object represents a final expression object (value, constant, variable,
# and others of IdObjects).
class Prim(Expr):
    # declared Signal, Constant, Parameter or SignalEvent or an int value
    prim=None

    # string representing prim
    str=''

    def __setattr__(self, key, value):
        if key=='prim':
            if value==None:
                self.__dict__['prim']=value
            elif EvalObject in type(value).mro():
                self.__dict__['prim']=value
            elif type(value)==str:
                self.__dict__['prim']=[c for c in value]
            elif type(value)==int:
                self.__dict__['prim']=value
            else:
                raise ValueError('value=%s' % value)
        else:
            super(Expr, self).__setattr__(key, value)

    def __str__(self):
        if IdObject in type(self.prim).mro():
            return self.prim.id
        else:
            return str(self.prim)
    
    # returns a prim
    def get_list(self):
        return self.prim


# expression generator function
def not_expr(expr):
    "not_expr(expr) returns OpExpr applying 'not' operator on expression"
    return OpExpr(operator='not', operands=[expr])

def and_expr(expr_list):
    "and_expr(expr_list) returns OpExpr applying 'and' on expresion list"
    return OpExpr(operator='and', operands=expr_list)

def or_expr(expr_list):
    "or_expr(expr_list) returns OpExpr applying 'or' on expresion list"
    return OpExpr(operator='or', operands=expr_list)
