#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:expandtab:sw=4:tw=0
###############################################################################
# simplify.py - simplify module for expression (logic or mathematical)        #
#                                                                             #
# $Id: simplify.py 201 2007-06-10 00:54:39Z smrcka $
#                                                                             #
# Author: Ales Smrcka <smrcka@fit.vutbr.cz>                                   #
#                                                                             #
# This program is free software; you can redistribute it and/or modify        #
#   it under the terms of the GNU General Public License as published by      #
#   the Free Software Foundation; either version 2 of the License, or         #
#   (at your option) any later version.                                       #
#   See http://www.gnu.org/licenses/gpl.html for more details.                #
#                                                                             #
###############################################################################

###############################################################################
# allowed operands (expressions):
#   a) int value (bool is represented by int 0=False, 1=True)
#   b) VariableUse object (or other object - only for get_used_objects())
#   c) list of [operator, operand1, operand2, ...]
# allowed operators (string type):
#   a) binary (n-ary): 
#       and, or, (associative)
#       ==, !=, <, >, <=, >=, (relational - has inverse operator)
#       +, *, (associative)
#   b) unary:
#       not (inverse),
#       neg (inverse),
#       abs (onetime - abs(abs(i))=abs(i))
###############################################################################

assoc_bool_operators=['and','or',]
assoc_int_operators=['+','*']
inverse_operators=['not','neg']
onetime_operators=['abs']
unary_operators = inverse_operators + \
                  onetime_operators

associative_operators = assoc_bool_operators + \
                        assoc_int_operators

_not_omega = {
    '==':'!=',
    '>':'<=',
    '<':'>=',
    '!=':'==',
    '<=':'>',
    '>=':'<',
    }

relational_operators=_not_omega.keys()


###############################################################################
# Variable class and some functions over expressions
###############################################################################

# a generic structure for temporary objects
class tmp(object):
    pass

# undefined value object
class Variable(object):
    # constuctor:
    # input: name - string literal
    #        type - string literal:
    #           bool = {0,1} (size=1)
    #           vector = {0,1,...,2^size-1} (size must be set)
    #           int = {-INF, +INF} (if size is unset) or
    #                 {-2^(size-1), +2^(size+1)-1} (is size is set)
    def __init__(self, name='', type='', size=-1):
        self.name=name
        self.type=type
        if self.type=='bool':
            self.size=1
        else:
            self.size=size
        self.flag = ''
        self.role = ''

    # automatic settings of size for bool type
    def __setattr__(self, key, value):
        if key=='type':
            self.__dict__[key]=value
            if value=='bool': self.size=1
        else:
            self.__dict__[key]=value

    # returns a string representative
    def __str__(self):
        return self.name

    # returns a shallow copy of this object
    def copy(self):
        var=Variable()
        for k,v in self.__dict__.iteritems():
            var.__setattr__(k,v)
        return var

    # returns xrange of all possible values
    def values(self):
        if self.size<0:
            if self.type=='int':
                raise ArithmeticError('infinite type')
            else:
                raise ArithmeticError('invalid type size')
        if self.type=='int':
            return xrange(-2**(self.size-1), +2**(self.size-1))
        else:
            return xrange(2**size)

    # returns a value, if defined (by name) in dictionary
    #        or self
    def evaluate(self, var_values):
        if var_values.has_key(self.name):
            return var_values[self.name]
        else:
            return self


# use of variable (may be: normal variable value, pos/neg edge, last)
class VariableUse(object):
    def __init__(self, v=None, use='normal'):
        # simplify.Variable reference
        self.v=v

        # type of use:
        # 'normal' - normal value of variable (s[v.name])
        # 'posedge' - not s[v.name] and s[v.name+'@next']
        # 'negedge' - s[v.name] and not s[v.name+'@next']
        # 'last' - s[v.name]
        # 'next' - s[v.name+'@next']
        # 'edge' - s[v.name] != s[v.name+'@next']
        # 'stable' - s[v.name] == s[v.name+'@next']
        # 'whatever' - s[v.name+'@whatever'] - universal one
        self.use=use

    def __getattr__(self, key):
        if key in self.__dict__.keys():
            return self.__dict__[key]
        elif key=='type':
            if self.isedge():
                return 'bool'
            else:
                return self.v.type

    # important for apply_equivalences(), for simplification - if two
    # different objects of VariableUse represents the same behavior,
    # for generating the set of objects
    def __eq__(self, vu):
        return type(vu)==type(self) and \
            ((self.v==vu.v) and (self.use==vu.use))
    def __hash__(self):
        return self.v.__hash__()^self.use.__hash__()

    def __str__(self):
        if self.use=='normal':
            return self.v.name
        else:
            return '%s@%s' % (self.v.name, self.use)

    def copy(self):
        return VariableUse(self.v, self.use)
    
    # list of edge variable use (static due to optimization)
    _edge_uses=['posedge','negedge']
    # returns True if self.use is either 'posedge' or 'negedge'
    def isedge(self):
        return self.use in self._edge_uses
        
    def evaluate(self, var_values):
        def v(name):
            if var_values.has_key(name):
                return var_values[name]
            else:
                return self
        if self.use=='normal':
            return v(self.v.name)
        elif self.use=='last':
            return v(self.v.name)
        elif self.use=='next':
            return v(self.v.name+'@next')
        elif self.use=='posedge':
            v_last=v(self.v.name)
            v_next=v(self.v.name+'@next')
            if self in [v_last, v_next]:
                return self
            else:
                assert (type(v_last)==int and type(v_next)==int)
                return (1-v_last) & v_next
        elif self.use=='negedge':
            v_last=v(self.v.name)
            v_next=v(self.v.name+'@next')
            if self in [v_last, v_next]:
                return self
            else:
                assert (type(v_last)==int and type(v_next)==int)
                return v_last & (1-v_next)
        elif self.use=='edge':
            v_last=v(self.v.name)
            v_next=v(self.v.name+'@next')
            if self in [v_last, v_next]:
                return self
            else:
                assert (type(v_last)==int and type(v_next)==int)
                return int(v_last!=v_next)
        elif self.use=='stable':
            v_last=v(self.v.name)
            v_next=v(self.v.name+'@next')
            if self in [v_last, v_next]:
                return self
            else:
                assert (type(v_last)==int and type(v_next)==int)
                return int(v_last==v_next)
        else:
            return v('%s@%s' % (self.v.name,self.use))
    
###############################################################################

# returns expr->text
def strexpr(expr):
    if type(expr)==int or expr==None:
        return str(expr)
    elif type(expr)==list:
        return '(%s)' % ','.join([expr[0]]+[strexpr(o) for o in expr[1:]])
    else:
        assert type(expr)==VariableUse
        return str(expr)

# returns a list of used objects (not int or list) within expression
# input: expression
# output: set of of used object
def get_used_object_list(expr, debug=0):
    if debug:
        print expr
    if type(expr)==int or expr==None:
        return []
    elif type(expr)==list:
        result=[]
        for i in expr[1:]:
            result+=get_used_object_list(i)
        return result
    else: # other type (used object)
        return [expr]

# returns a set of used objects (not int or list) within expression
# input: expression
# output: set of of used object
def get_used_objects(expr):
    if type(expr)==int or expr==None:
        return set()
    elif type(expr)==list:
        result=set()
        for i in expr[1:]:
            result|=get_used_objects(i)
        return result
    else: # other type (used object)
        return set([expr])

# returns a set of used variables within expression
# input: expression
# output: set of Variable objects
def get_variables(expr):
    if type(expr)==int or expr==None:
        return set()
    elif type(expr)==list:
        result=set()
        for i in expr[1:]:
            result|=get_variables(i)
        return result
    else: # VariableUse type
        assert type(expr)==VariableUse
        return set([expr.v])

# returns a shallow copy of an expression
def copy_expression(expr):
    if type(expr)==int:
        return expr
    elif type(expr)==list:
        result=[expr[0]]
        for i in expr[1:]:
            result.append(copy_expression(i))
        return result
    else: # VariableUse type
        assert type(expr)==VariableUse
        return expr.copy()

# tries to evaluate an expression in order to variableuse values dictionary
# input: expression, variableuse values dictionary (id (or name) -> value)
#        allow_var - if Variable type is allowed
# output: simplified expression of VariableUse (if not all variables have
#           their value)
#         int (resulting value)
def evaluate(expr, values, allow_var=False):
    if type(expr)==int or expr==None:
        return expr
    elif type(expr)==VariableUse:
        return expr.evaluate(values)
    elif type(expr)==Variable and allow_var:
        return expr.evaluate(values)
    # expr is a list [operator, operand1, operand2, ...]
    elif type(expr)!=list:
        print 'Its evaluate: expr=%s' % expr
    #assert type(expr)==list
    operator=expr[0]
    # evaluate all of operands
    operands=[evaluate(e, values) for e in expr[1:]]
    # if some VariableUse remains, return a simplified expression...
    for e in operands:
        if type(e) in [list,VariableUse]:
            e=[operator]+operands
            return apply_equivalences(e)
    # ...othervise evaluate an expression
    if operator=='and':
        result=1
        for o in operands:
            result=result and o
        return int(result)
    elif operator=='or':
        result=0
        for o in operands:
            result=result or o
        return int(result)
    elif operator=='not':
        return int(not operands[0])
    elif operator=='==':
        result=operands[0]
        for o in operands:
            result=result and (result==o)
        return int(result)
    elif operator=='!=':
        return int(int(operands[0]) != int(operands[1]))
    elif operator=='>':
        return int(int(operands[0]) > int(operands[1]))
    elif operator=='<':
        return int(int(operands[0]) < int(operands[1]))
    elif operator=='>=':
        return int(int(operands[0]) >= int(operands[1]))
    elif operator=='<=':
        return int(int(operands[0]) <= int(operands[1]))
    elif operator=='+':
        result=0
        for o in operands:
            result+=o
        return result
    elif operator=='*':
        result=0
        for o in operands:
            result*=int(o)
        return result
    elif operator=='-':
        return int(operands[0]) - int(operands[1])
    elif operator=='abs':
        return abs(int(operands[0]))
    elif operator=='div':
        return int(operands[0]) / int(operands[1])
    elif operator=='mod':
        return int(operands[0]) % int(operands[1])
    
def substitute_use(expr1, var, expr2):
    """substitutes every use of variable var in expr1 with expr2
    input: expr1, var, expr2
    output: substituted expr1"""
    if type(expr1)==list:
        for i in xrange(1, len(expr1)):
            expr1[i]=substitute_use(expr1[i], var, expr2)
        return expr1
    else:
        assert type(expr1) in [int,VariableUse]
        if type(expr1)==VariableUse and expr1.v==var:
            return expr2
        else:
            return expr1

###############################################################################
# Expression simplification
###############################################################################

# Shortening by expression equivalences
def apply_equivalences(expr):
    def r0(expr):
        # omega,o; omega ∉ unary_operators -> o
        if expr[0] in unary_operators: return (False, expr)
        if len(expr)==2:
            return (True, expr[1])
        return (False, expr)
    def r1(expr):
        # and,o1,o2,0,... -> 0
        if expr[0]!='and': return (False, expr)
        for i in expr[1:]:
            if not i:
                return (True, 0)
        return (False, expr)
    def r2(expr):
        # and,o1,o2,1,1,... -> and,o1,o2,...
        if expr[0]!='and': return (False, expr)
        changed=False
        e=[expr[0]]
        for i in expr[1:]:
            if i==1:
                changed=True
            else:
                e.append(i)
        if changed:
            return (True, e)
        return (False, expr)
    def r3(expr):
        # and,o1,o2,o1,o3,... -> and,o1,o2,o3,...
        if expr[0]!='and': return (False, expr)
        e=[]
        for o in expr[1:]:
            if o not in e:
                e.append(o)
        e.insert(0, expr[0])
        if e!=expr:
            return (True, e)
        return (False, expr)
    def r4(expr):
        # and,o1,not o1,... -> 0
        if expr[0]!='and': return (False, expr)
        e=expr[1:]
        for o in e:
            if ['not', o] in e:
                return (True, 0)
        return (False, expr)
    def r5(expr):
        # or,o1,o2,1,... -> 1
        if expr[0]!='or': return (False, expr)
        for i in expr[1:]:
            if i==1:
                return (True, 1)
        return (False, expr)
    def r6(expr):
        # or,o1,o2,0,0,... -> or,o1,o2,...
        if expr[0]!='or': return (False, expr)
        changed=False
        e=[expr[0]]
        for i in expr[1:]:
            if i==0:
                changed=True
            else:
                e.append(i)
        if changed:
            return (True, e)
        return (False, expr)
    def r7(expr):
        # or,o1,o2,o1,o3,... -> or,o1,o2,o3,...
        if expr[0]!='or': return (False, expr)
        e=[]
        for o in expr[1:]:
            if o not in e:
                e.append(o)
        e.insert(0, expr[0])
        if e!=expr:
            return (True, e)
        return (False, expr)
    def r8(expr):
        # or,o1,not o1,... -> 1
        if expr[0]!='or': return (False, expr)
        e=expr[1:]
        for o in e:
            if ['not', o] in e:
                return (True, 1)
        return (False, expr)
    def r9(expr):
        # not,(omega,A,B); omega ∈ relational_operators -> not_omega,A,B
        if expr[0]!='not': return (False, expr)
        if type(expr[1])==list and len(expr[1])==3:
            omega=expr[1][0]
            if omega in _not_omega.keys():
                return (True, [_not_omega[omega], expr[1][1], expr[1][2]])
        return (False, expr)
    def r10(expr):
        # omega,(omega,A); omega ∈ inverse_operators -> A
        if expr[0] not in inverse_operators: return (False, expr)
        if type(expr[1])==list and expr[0]==expr[1][0]:
            return (True, expr[1][1])
        return (False, expr)
    def r11(expr):
        # ==,A,1; A is bool -> A
        if expr[0]!='==': return (False, expr)
        if len(expr)==3:
            if expr[1]==1: i=2
            elif expr[2]==1: i=1
            else: return (False, expr)
            if expr[i].type=='bool':
                return (True, expr[i])
            try:
                if expr[i].type=='bool':
                    return (True, expr[i])
            except AttributeError:
                raise AttributeError('type(expr)=%s, value=%s' % \
                    (type(expr), expr))
        return (False, expr)
    def r12(expr):
        # ==,A,0; A is bool -> (not,A)
        if expr[0]!='==': return (False, expr)
        if len(expr)==3:
            if expr[1]==0: i=2
            elif expr[2]==0: i=1
            else: return (False, expr)
            try:
                if expr[i].type=='bool':
                    return (True, ['not', expr[i]])
            except AttributeError:
                pass
        return (False, expr)
    def r13(expr):
        # !=,A,1; A is bool -> (not,A)
        if expr[0]!='!=': return (False, expr)
        if len(expr)==3:
            if expr[1]==1: i=2
            elif expr[2]==1: i=1
            else: return (False, expr)
            try:
                if expr[i].type=='bool':
                    return (True, ['not', expr[i]])
            except AttributeError:
                pass
        return (False, expr)
    def r14(expr):
        # !=,A,0; A is bool -> A
        if expr[0]!='!=': return (False, expr)
        if len(expr)==3:
            if expr[1]==0: i=2
            elif expr[2]==0: i=1
            else: return (False, expr)
            try:
                if expr[i].type=='bool':
                    return (True, expr[i])
            except AttributeError:
                pass
        return (False, expr)
    def r15(expr):
        # omega,A,(omega,B1,B2),...; omega ∈ associative_operators ->
        #       omega,A,B1,B2,...
        if expr[0] not in associative_operators: return (False, expr)
        omega=expr[0]
        for i in xrange(1,len(expr)):
            if type(expr[i])==list and expr[i][0]==omega:
                return (True, [omega]+expr[1:i]+expr[i][1:]+expr[i+1:])
        return (False, expr)
    def r16(expr):
        # and,A,(or,A,B) -> A
        if expr[0]!='and': return (False, expr)
        if len(expr)==3:
            if type(expr[1])==list and type(expr[2])!=list:
                l=1 #list idx
                A=2 #A idx
            elif type(expr[2])==list and type(expr[2])==list:
                l=2 #list idx
                A=1 #A idx
            else: return (False, expr)
            if expr[l][0]=='or':
                if expr[A] in expr[l][1:]:
                    return (True, expr[A])
        return (False, expr)
    def r17(expr):
        # or,A,(and,A,B) -> A
        if expr[0]!='or': return (False, expr)
        if len(expr)==3:
            if type(expr[1])==list and type(expr[2])!=list:
                l=1 #list idx
                A=2 #A idx
            elif type(expr[2])==list and type(expr[2])==list:
                l=2 #list idx
                A=1 #A idx
            else: return (False, expr)
            if expr[l][0]=='and':
                if expr[A] in expr[l][1:]:
                    return (True, expr[A])
        return (False, expr)
    def r18(expr):
        # omega,(omega,A); omega ∈ onetime_operators -> (omega,A)
        if expr[0] not in onetime_operators: return (False, expr)
        if type(expr[1])==list and expr[0]==expr[1][0]:
            return (True, expr[1])
        return (False, expr)
    def r19(expr):
        # not,0 -> 1
        # not,1 -> 0
        if expr[0]!='not': return (False, expr)
        if expr[1]==0: return (True, 1)
        if expr[1]==1: return (True, 0)
        return (False, expr)
    def r20(expr):
        # and what about?
        # (or,(and,A,B,C...),(and,A,(not,B),C...),Z) -> (or,(and,A,C...),Z)
        pass

    if type(expr)==list:
        for i in xrange(1,len(expr)):
            expr[i]=apply_equivalences(expr[i])
    changed=True
    rules=[r0,r1,r2,r3,r4,r5,r6,r7,r8,r9,r10, \
           r11,r12,r13,r14,r15,r16,r17,r18,r19]
    while changed and type(expr)==list:
        for r in rules:
            changed,expr=r(expr)
            if changed: break
    return expr

###############################################################################
