#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:expandtab:sw=4:tw=0
###############################################################################
# vhd2ca.py - hardware description to counter automaton transformation        #
#                                                                             #
# $Id: vhd2ca.py 362 2007-11-13 14:16:21Z smrcka $
#                                                                             #
# Author: Ales Smrcka <smrcka@fit.vutbr.cz>                                   #
#                                                                             #
# This program is free software; you can redistribute it and/or modify        #
#   it under the terms of the GNU General Public License as published by      #
#   the Free Software Foundation; either version 2 of the License, or         #
#   (at your option) any later version.                                       #
#   See http://www.gnu.org/licenses/gpl.html for more details.                #
#                                                                             #
###############################################################################

import sys  # argv, stdout, stderr

# global options 
OPTIONS = {
    'verbose': 1,
    'color': 1,
    'generate': 1,
    'excludes': {},
    'constants': {},
    'inits': {},
    'badvar': '',
    'output type': '',
    'output': '',
    'output options': {},
}

APPPATH=''

###############################################################################
# General functions and properties
###############################################################################
class TransformationError(Exception):
    pass

# System name to be verified (from the first source file)
system_name='Noname'

# After parsing, entity stands for Entity object to be verified.
entity=None

# a shortcut for raising errors
def t_error(msg):
    raise TransformationError(msg)

# returns text representation of simple state transition
def str_two_states(s):
    names=[name for name in s.keys() if name.find('@')<0]
    names.sort()
    result=[]
    for name in names:
        result.append('%s:%s->%s' % (colorstr(name, 'red'),
            s[name], s[name+'@next']))
    return '  '.join(result)

###############################################################################
# Some unimportant utilites
###############################################################################

current_progress=None
# Progress bar - algorithm is very slow, for the user not to go mad
class Progress(object):
    def __init__(self, msg, max, output=sys.stdout, width=30, verbose=1):
        """create a text progress bar
        msg - text message before progress
        max - maximal items to step
        output - output file
        width - progress bar width in characters
        verbose - """
        self.msg=msg
        self.max=max
        self.curr=0
        self.output=output
        self.width=width
        self.verbose=verbose
        self.running=True
        self.printprogress(self.fmt_msg())
        global current_progress
        current_progress=self

    def next(self, inc=1):
        self.running=True
        self.curr+=inc
        if self.curr>=self.max:
            self.done()
        else:
            c=self.fmt_msg()
            if self.currmsg!=c: self.printprogress(c)

    def done(self):
        self.printprogress('%s [%s] 100%%\n' % (self.msg, '#'*self.width))
        global current_progress
        current_progress=None
        self.running=False
    
    def pause(self, file=sys.stdout):
        if self.running and file==self.output:
            self.output.write('\n')
            self.output.flush()
        self.running=False

    def stop(self):
        self.running=False
        if self.curr<self.max:
            self.printprogress('%s - incomplete\n' % self.currmsg)
            global current_progress
            current_progress=None

    def printprogress(self, c):
        self.currmsg=c
        if OPTIONS['verbose']>=self.verbose and OPTIONS['verbose']<4:
            self.output.write('\r')
            self.output.write(c)
            self.output.flush()

    def fmt_msg(self):
        i=self.width*self.curr/self.max
        return '%s [%s%s] %d%%' % (self.msg, \
            '#'*i, ' '*(self.width-i), 100*self.curr/self.max)

# playing with colors
text_color={
    'init': '\x1b[!p\x1b[?3;4l\x1b[4l\x1b>',
    'default': '\x1b[0m',
    '': '\x1b[0m',
    'bold': '\x1b[1m',
    'black': '\x1b[30m',
    'red': '\x1b[31m',
    'green': '\x1b[32m',
    'brown': '\x1b[33m',
    'blue': '\x1b[34m',
    'magenta': '\x1b[35m',
    'cyan': '\x1b[36m',
    'gray': '\x1b[37m',
}

def set_text_color(color):
    if OPTIONS['color']: sys.stdout.write(text_color[color])

def color(c=''):
    if OPTIONS['color']: return text_color[c]
    else: return ''

def colorstr(s, c=''):
    return '%s%s%s' % (color(c),s,color('default'))

def colorbehavior(b):
    return '%s -> %s:=%s' % (colorstr(strexpr(b.guard), 'blue'),
        colorstr(b.var.name, 'red'),
        colorstr(strexpr(b.value), 'brown'))

###############################################################################
# Classes for states, transitions and counter automata
###############################################################################

# state representative; includes dictionary id->value (s) of state property
# evaluations
class State(object):
    # creates a state from variable evaluation v (i.e., dict of id->value)
    def __init__(self, v, id):
        self.v=v
        self.id=id

    # string representation of state
    def __str__(self):
        return ', '.join(['%s=%s' % (k,v) for k,v in self.v.iteritems()])

    def is_bad(self):
        "returns 1 if contains true bad variable evaluation"
        badvar_name = OPTIONS['badvar']
        if self.v.has_key(badvar_name):
            if self.v[OPTIONS['badvar']]:
                return 1
            else:
                return 0
        else:
            return 0

# action of a transition
class Action(object):
    # constructor of action
    # input: counter - Variable object
    #        value - expression
    def __init__(self, counter, value):
        self.counter=counter
        self.value=value

    def __str__(self):
        return '%s := %s' % (colorstr(self.counter.name, 'red'), \
            colorstr(strexpr(self.value), 'brown'))

    def __eq__(self, a):
        return type(a)==Action and \
               self.counter==a.counter and self.value==a.value

non_action = None

# transition between two states
class Transition(object):
    # constructor of transition:
    # input: s1 ∈ states - FROM state
    #        s2 ∈ state - TO state
    #        guard - expression
    #        action - one Action (counter <- value)
    def __init__(self, s1=None, s2=None, guard=None, action=None):
        self.s1=s1
        self.s2=s2
        self.guard=guard
        self.action=action
        self.actions=[]

    def __str__(self):
        if self.s1 and self.s2:
            trans='%s->%s: ' % (self.s1.id, self.s2.id)
        else:
            trans=''
        if self.actions:
            return '%s%s; %s' % (trans, \
                colorstr(strexpr(self.guard), 'blue'), \
                ', '.join([str(a) for a in self.actions]))
        if self.action:
            return '%s%s; %s' % (trans, \
                colorstr(strexpr(self.guard), 'blue'), self.action)
        else:
            return '%s%s' % (trans, colorstr(strexpr(self.guard), 'blue'))

    def __eq__(self, t):
        return type(t)==Transition and \
               self.s1==t.s1 and \
               self.s2==t.s2 and \
               self.guard==t.guard and \
               self.action==t.action

# list objects are not hashable
def list_hash(l):
    if type(l)==list:
        hash=0
        for i in l:
            hash = hash ^ (i<<1)
        return hash
    else:
        return l.__hash__()

# do step 5e) from paper notes
def group_transition_actions(transitions):
    # make set of transitions
    # little complicated due to unable to hash the list (expression),
    # equals to: ts = set(transition)
    ts=[]
    for t in transitions:
        if t not in ts: ts.append(t)
    ts=set(ts)
    # group actions over the same guard
        # find unique guards: guards = list_from( { guard(t) | t \in ts } )
    guards=[]
    for t in ts:
        if t.guard not in guards: guards.append(t.guard)
        # group actions
    # conditionals are guards that condition the whole transition,
    # such rules set 1-bit state variable (has no action, because the
    # transition itself provides such an assignment).
    conditionals=[]
    for t in ts:
        if t.action==None: conditionals.append(t.guard)
    # conditional is conditional expression in conjunction
    if len(conditionals)>1:
        conditional=['and']+conditionals
    elif len(conditionals)==1:
        conditional=conditionals[0]
    else:
        conditional=1
    # group actions over guard
    guards=[[g,[]] for g in guards] # g[0]=guard, g[1]=[Transition]
    for i in xrange(len(guards)):
        for t in ts:
            if t.guard==guards[i][0]: guards[i][1].append(t)
    # let g[x][0] be Transition object with .actions=[g[x][1:].action]
    for g in guards:
        if g[1][0].action:
            g[1][0].actions=[g[1][0].action]
        else:
            g[1][0].actions=[]
        for t in g[1][1:]:
            if t.action:
                g[1][0].actions.append(t.action)
    ts=[g[1][0] for g in guards] # ts=[Transition with .actions=[expr]]
    # we have 3 types of transitions:
    #   1. guard==1: always perfomed actions
    #   2. guard in conditionals: every transition must cover such a guard
    #   3. guard!=1 and not in conditionals: partial guard,
    #                       corresponding action is possible (non1guards)
    # make resulting transitions:
    non1guards=[t for t in ts \
        if (t.guard!=1) and (t.guard not in conditionals)]
    nguards=len(non1guards)

    new_ts=[]
    if nguards>1:
        for iter in xrange(1,2**nguards): # [and,not g1,not g2] is not allowed
            # guard will be [and,g1,[not,g2]] or ... or [and,g1,g2]
            guard=['and']
            # actions will be list of possible actions 
            actions=[]
            for i in xrange(nguards):
                if (iter>>i)&1:
                    guard.append(non1guards[i].guard)
                    actions+=non1guards[i].actions
                else:
                    guard.append(['not',non1guards[i].guard])
            t=Transition(non1guards[i].s1, non1guards[i].s2, guard, None)
            t.actions=actions
            new_ts.append(t)
    elif nguards==1:
        new_ts=non1guards
    trueguard=[t for t in ts if t.guard==1] # either one or none
    if trueguard:
        if new_ts:
            for t in new_ts:
                t.actions+=trueguard[0].actions
        else:
            new_ts=trueguard
    for t in new_ts:
        t.guard=evaluate(['and',t.guard,conditional], {})
    return new_ts

###############################################################################
# Counter Automaton + generating methods
class CounterAutomaton(object):

    # constructor
    def __init__(self):
        # name of a system
        self.name=''
        # set of State
        self.states=set()
        # initial state
        self.init_state=None
        # set of Transition
        self.transitions=[]
        # set of parser.EvalObject
        self.counters=set()
        # set of parser.Parameter
        self.parameters=set()
        # set of parser.EvalObject
        self.state_vars=set()
        # behavior over state variables, set of Behavior
        self.behavior=set()
        # set of synchronization problems
        self.syn_problems=set()
    
    # returns list of variable use, where variable is either state or counter
    def vu_list(self, expr):
        return [vu for vu in get_used_object_list(expr) \
            if (vu.v in self.state_vars) or (vu.v in self.counters)]

    ###########################################################################
    # state space generator; list of state variables is required
    # input: self.state_vars
    def generate_states(self):
        # generator of a new state;
        # i (int) has binary combination of ordered state variable (osv) values
        def new_state(i, osv):
            def tobinary(number, count):
                v=[]
                for i in xrange(count):
                    v.append(number&1)
                    number>>=1
                return v
            s=dict(zip(osv, tobinary(i, len(osv))))
            return State(s, 's%d' % i)
        self.states=[]
        ordered_state_vars = [v.name for v in list(self.state_vars)]
        ordered_state_vars.sort()
        top=2**len(ordered_state_vars)
        i=0
        while i<top:
            s=new_state(i, ordered_state_vars)
            self.states.append(s)
            if OPTIONS['verbose']>=3:
                print '%s: %s' % (colorstr('state %s' % s.id, 'bold'), s)
            i+=1

    ###########################################################################
    # generator of transitions between states; set of states is required
    # input: self.states, self.actions, self.state_vars
    def generate_transitions(self):
        # returns clockuse if clock event possibly sets variable@next value
        # input: b - behavior action over variable
        #        s - transition dictionary
        # output: VariableUse (edge-triggered) or 1 (transparent) or None
        #         (not set)
        #         b.optim.guard - from clocks or counters),
        #         b.optim.varuse - VariableUse list if guard!=0]
        def who_triggers(b, s):
            # evaluate all sensitivity list variables (up to clock edge)
            # as @next
            s_part={}
            for u in b.sensitive:
                if u.isedge(): break
                s_part[u.v.name]=s[u.v.name+'@next']

            # evaluate guard with some sensitivity variables
            b.optim.guard=evaluate(b.guard, s_part)
            # behavior still needs values of these variables:
            b.optim.varuse=get_used_objects(b.optim.guard)
            if b.optim.guard: # if action is possible...
                if b.optim.guard==1:
                    # action is 100% allowed (either true or transparent)
                    return 1
                else: # guard is an expression
                    clk=set([u for u in b.optim.varuse if u.isedge()])
                    if len(clk)==0:
                        return 1
                    elif len(clk)==1:
                        clk=list(clk)[0]
                        s_part[clk.v.name]=s[clk.v.name]
                        s_part[clk.v.name+'@next']=s[clk.v.name+'@next']
                        b.optim.guard=evaluate(b.optim.guard, s_part)
                        b.optim.varuse=get_used_objects(b.optim.guard)
                        if b.optim.guard:
                            return clk
                        else:
                            return None
                    else:
                        t_error('Unsupported sensitivity list for %s: %s' % \
                            (b.var.name, ','.join([c.name for c in clks])))
            else:
                return None # action is not allowed

        # evaluate every variable with None (not set), 1 (transparent), or
        #   clock use (edge-triggered)
        # input: s - state variable values with '' and '@next' values
        # output: variable.optim.trigger - who triggers it
        #         behaviors.optim.guard - linked part of behavior.guard for s
        #         behaviors.optim.varuse - used variableuse in linked guard
        def update_triggers(s):
            for v in self.variables:
                v.optim.trigger=None # not set at default
            for b in self.behavior:
                if b.var:
                    c=who_triggers(b, s)
                    if c:
                        assert (b.var.optim.trigger==None) or \
                               (b.var.optim.trigger==c)
                            # only one action(event) can change var. value
                        b.var.optim.trigger=c

        # g -> sv := value =>
        # output:
        #  1. g and sv@next and value -> None := None
        #  2. g and not sv@next and not value -> None := None
        def split_statevar_assignment():
            new_behavior=set()
            for b in self.behavior:
                if b.var in self.state_vars:
                    guard1=['and', copy_expression(b.guard),
                        VariableUse(b.var, 'next'), copy_expression(b.value)]
                    guard1=evaluate(guard1, {})
                    guard2=['and', copy_expression(b.guard),
                        ['not', VariableUse(b.var, 'next')],
                        ['not', copy_expression(b.value)]]
                    guard2=evaluate(guard2, {})
                    if guard1!=0:
                        b1=Behavior(guard1, b.var, None)
                        b1.sensitive=b.sensitive
                        b1.vhdl_source=b.vhdl_source
                        new_behavior.add(b1)
                    if guard2!=0:
                        b2=Behavior(guard2, b.var, None)
                        b2.sensitive=b.sensitive
                        b2.vhdl_source=b.vhdl_source
                        new_behavior.add(b2)
                else:
                    new_behavior.add(b)
            self.behavior=new_behavior

        # evaluate the behavior for transition s (s1 values -> s2 values)
        # input: b - behavior (with b.optim, b.var.optim)
        #        s - transition = state variable values
        #        newsyn - empty set for possible syn problems
        # output: list of Transition objects if transition is possible
        #         updated newsyn
        def evaluate_transition(b, s, newsyn):
            # output: list of s with v@rndsel values (for v in arbitr_var)
            def evaluation_list(s, arbitr_var):
                result=[s]
                for v in arbitr_var:
                    result_new=[]
                    for e in result:
                        e_new=e.copy()
                        e[v+'@rndsel']=e[v]
                        e_new[v+'@rndsel']=e[v+'@next']
                        result_new.append(e)
                        result_new.append(e_new)
                    result=result_new
                return result

            # mark every variable use as @next or @last (or @rndsel)
            # arbitr_var is updated set of state var. names with @rndsel use
            def mark_use(arbitr_var):
                if b.var.optim.trigger==None:
                    t_error('No next value specified for %s\n  %s' % \
                        (b.var.name, str_two_states(s)))
                elif b.var.optim.trigger==1: # transparent
                    for vu in self.vu_list(b.optim.guard)+ \
                              self.vu_list(b.value):
                        if vu.use=='normal': vu.use='next'
                else: # edge-triggered, VariableUse object
                    l=self.vu_list(b.optim.guard)+self.vu_list(b.value)
                    for vu in self.vu_list(b.optim.guard)+ \
                              self.vu_list(b.value):
                        if vu.use=='normal':
                            if vu.v.optim.trigger==b.var.optim.trigger:
                                vu.use='last'
                            else:
                                vu.use='rndsel' # select either last or next
                                if vu.v.flag!='counter':
                                    arbitr_var.add(vu.v.name)

            if not b.optim.guard: return []
            arbitr_var=set()
            mark_use(arbitr_var)
            transitions=set()
            for s in evaluation_list(s, arbitr_var):
                guard=evaluate(b.optim.guard, s)
                if guard:
                    value=evaluate(b.value, s)
                    # possible synchronization problem,
                    # enlist a warning TODO
                    rnds=[vu.v.name for vu in get_used_objects(guard)
                        if vu.use=='rndsel' and vu.v not in self.parameters]
                    rnds+=[vu.v.name for vu in get_used_objects(value)
                        if vu.use=='rndsel' and vu.v not in self.parameters]
                    rnds=set(rnds)
                    for name in rnds:
                        newsyn.add((name, b.var.name, b.vhdl_source))

                    if value==None:
                        action=None
                    else:
                        action=Action(b.var, value)
                    t=Transition(guard=guard, action=action)
                    transitions.add(t)
            return transitions 

        # check for missing variable value specification
        no_specification=[]
        for v in self.variables:
            if self.var_spec[v]==[]:
                no_specification.append(v.name)
        if no_specification:
            t_error('Variable values specification missing: %s' % \
                    ','.join(no_specification))
        # for every variable, prepare store for optimization info
        for v in self.variables:
            v.optim=tmp()
        # split: guard -> sv := value (value=1 (sv@next) or 0 (not sv@next))
        split_statevar_assignment()
        # save original use of every variable use
        for b in self.behavior:
            for vu in get_used_object_list(b.guard):
                vu.origuse=vu.use
            for vu in get_used_object_list(b.value):
                vu.origuse=vu.use
        # generate simple transitions
#        states=self.states[:8]
        states=self.states
        progress=Progress('Generating transitions', len(states))
        # set of synchronization problems
        self.syn_problems=set()
        for s1 in states:
            progress.next()
            if s1.is_bad(): # do not generate transitions FROM bad state
                continue
            for s2 in states:
                if s1!=s2: # for ∀s1,s2∈S, s1!=s2:
                    # restore original variable use
                    for b in self.behavior:
                        for vu in get_used_object_list(b.guard):
                            vu.use=vu.origuse
                        for vu in get_used_object_list(b.value):
                            vu.use=vu.origuse
                    # s = s1->s2
                    s=s1.v.copy()
                    for k,v in s2.v.iteritems():
                        s[k+'@next']=v
                    # prepare b.optim
                    for b in self.behavior:
                        b.optim=tmp()
                    # setup `trigger' for every variable set by behavior
                    # step 5 a) (from paper notes)
                    update_triggers(s)
                    # steps 5 b)-d) (from paper notes)

                    # every state variable value must be committed to allow
                    # this transition
                    for v in self.state_vars:
                        v.optim.commit=False
                    transitions=set()
                    newsyn=set() # space for possible syn problems
                    for b in self.behavior:
                        for t in evaluate_transition(b, s, newsyn):
                            t.s1=s1
                            if s2.is_bad():
                                t.s2=self.bad_state
                            else:
                                t.s2=s2
                            transitions.add(t)
                            b.var.optim.commit=True
                    # is every state variable value variable committed?
                    commit=reduce(lambda x,y: x and y, \
                            [v.optim.commit for v in self.state_vars])
                    if commit:
                        # update synchronization problems
                        self.syn_problems |= newsyn
                        # union a transitions
                        transitions=group_transition_actions(transitions)
                        # add new transitions to counter automaton
                        for t in transitions:
                            if t not in self.transitions:
                                self.transitions.append(t)
                        if OPTIONS['verbose']>=4:
                            for t in transitions:
                                print '%s: %s' % \
                                    (colorstr('Next transition', 'bold'), t)
                        if OPTIONS['verbose']>=5: print str_two_states(s)
        # add transitions to bad state (if bad variable is nonstate)
        if CA.badvar and CA.badvar.flag == 'nonstate':
            for s in self.states:
                guard = evaluate(CA.badvar.value, s.v)
                if guard:
                    t = Transition(s, CA.bad_state, guard, None)
                    self.transitions.append(t)
    
CA=CounterAutomaton()

###############################################################################
# Variable and counters filters (from VHDL entity)
###############################################################################

# Finding state variables. State variables are:
#   a) one bit input ports
#   b) one bit signals which has some non-true guard before assignment
# input: variables (ad a))
#        behavior (ad b))
# output: state variables marked as 'state'
def mark_state_variables(variables, behavior):
    # a) variables
    for v in variables:
        if v.type=='bool' and \
           type(v.vhdl)==parser.Port and \
           v.vhdl.direction in ['in','inout']:
            if not v.flag.startswith('state'):
                v.flag='state'
    # b) variables
    for b in behavior: # a set of Behavior(guard,dest,value)
        # guard other than (1,dest,value) and bool type variables
        if b.guard!=1 and b.var.type=='bool':
            if not b.var.flag.startswith('state'):
                b.var.flag='state'

# A test for cyclic assignment in Action,
# TransformationError is raised if cycle detected.
# input: behavior (one Behavior(guard,dest,value))
def test_cyclic_assign(behavior):
    if behavior.var.flag=='nonstate':
        if behavior.var in get_variables(behavior.value):
            t_error('cyclic assignment detected: %s' % behavior)

# Finding nonstate variables. Nonstate variables are signals which has some
# true guard before assignment.
# input: variables, behavior
# output: nonstate variables marked as 'nonstate'
def mark_nonstate_variables(variables, behavior):
    for b in behavior:
        if b.guard==1:
            b.var.flag='nonstate'
            test_cyclic_assign(b)

# Finding counters. Counters are more bit state variables.
# input: variables
#        behavior
# output: counters variables marked as 'counter'
def mark_counters(variables, behavior):
    for v in variables:
        if v.type=='int' and \
           type(v.vhdl)==parser.Port and \
           v.vhdl.direction in ['in','inout']:
            v.flag='counter'
    for b in behavior:
        if b.guard!=1 and b.var.type=='int':
            b.var.flag='counter'

# Finding parameters.
# input: variables
#        behavior (not used)
# output: parameter variables marked as 'parameter'
def mark_parameters(variables, behavior):
    for v in variables:
        if type(v.vhdl)==parser.Parameter:
            v.flag='parameter'

# Substitutes all of nonstate variables in expressions for expression over
# state variables.
# input: variables, behavior
# output: none (changes behavior body)
def substitute_nonstate_variables(variables, behavior):
    def subst_in_sensitive(sensitive, variable, new_sensitive):
        changed=True
        while changed:
            changed=False
            for i in xrange(len(sensitive)):
                if sensitive[i].v.name==variable.name:
                    sensitive=sensitive[:i]+new_sensitive+sensitive[i+1:]
                    changed=True
                    break
        return sensitive

    nonstate_behavior = set([b for b in behavior \
        if b.var.flag == 'nonstate' and b.var.role != 'bad'])
    for b1 in nonstate_behavior:
        behavior-=set([b1])
        for b2 in behavior:
            b2.guard=substitute_use(b2.guard, b1.var, b1.value)
            b2.value=substitute_use(b2.value, b1.var, b1.value)
            sl=[VariableUse(v) for v in get_variables(b1.value)]
            b2.sensitive=subst_in_sensitive(b2.sensitive, b1.var, sl)
            test_cyclic_assign(b2)
            b2.guard = evaluate(b2.guard, {})
            b2.value = evaluate(b2.value, {})

# shorten guards and values for faster evaluation, preserve only state
# variables in sensitivity list
# input: behavior
# output: changed behavior (insitu)
def cleared_behavior(behavior):
    for b in behavior:
        b.guard=evaluate(b.guard, {})
        b.value=evaluate(b.value, {})
        sensitive=b.sensitive
        b.sensitive=[]
        for s in sensitive:
            if s.v.flag in ['state','state clock']:
                b.sensitive.append(s)
    return behavior

# generate CA.var_spec dictionary (variable->[behaviors setting variable])
def update_var_spec(CA):
    # partition behaviors wrt variables
    CA.var_spec={}
    for b in CA.behavior:
        b.guard=evaluate(b.guard, {})
        b.value=evaluate(b.value, {})
        if b.guard:
            if CA.var_spec.has_key(b.var):
                CA.var_spec[b.var].append(b)
            else:
                CA.var_spec[b.var]=[b]

def update_used_vars(CA):
    "moves variables with false guards only to temp_vars"
    new_temp_vars = set()
    for var in list(CA.state_vars):
        if var not in CA.var_spec.keys():
            var.flag = 'nonstate'
            CA.temp_vars[var] = 'undefined'
            CA.state_vars.remove(var)
            new_temp_vars.add(var)
    for var in list(CA.counters):
        if var not in CA.var_spec.keys():
            var.flag = 'nonstate'
            CA.temp_vars[var] = 'undefined'
            CA.counters.remove(var)
            new_temp_vars.add(var)
    # if environment (user specified constants) causes to invariant value
    # of internal variable (all guard are false), we must use its initial
    # value.
    init_state = CA.init_transition.s1
    for var in new_temp_vars:
        if var.name not in OPTIONS['constants']:
            # find its init value (var_init)
            if var.name in init_state.v.keys():
                var_init = init_state.v[var.name]
            else:
                var_init = [a for a in CA.init_transition.actions \
                    if a.counter == var]
                var_init = var_init[0].value
            # substitute in every used behavior
            for behs in CA.var_spec.values():
                for beh in behs:
                    beh.guard = substitute_use(beh.guard, var, var_init)
                    beh.value = substitute_use(beh.value, var, var_init)
                    beh.sensitive = [vu for vu in beh.sensitive if vu.v!=var]
    # update CA.{behavior,variables,used variables in init_transition}
    CA.behavior = set()
    for behs in CA.var_spec.values():
        CA.behavior |= set(behs)
    CA.variables = set(CA.var_spec.keys())
    CA.init_transition.actions = [a for a in CA.init_transition.actions \
        if a.counter in CA.variables]

# add behavior to preserve same value of variable if no guard is active
def add_same_value_behavior(CA):
    update_var_spec(CA)

    # append new same value behavior
    for v,behavior in CA.var_spec.iteritems():
        if v in CA.state_vars:
            guard=['and']
            sensitive=[]
            for b in behavior:
                guard.append(['not',copy_expression(b.guard)])
                for s in b.sensitive:
                    if s not in sensitive: sensitive.append(s)
            guard=evaluate(guard, {})
            b=Behavior(guard, v, VariableUse(v, 'last'))
            b.sensitive=sensitive
            CA.behavior.add(b)
            CA.var_spec[v].append(b)

# returns set of variable possibly representing transparent reset signal
def find_resets(behavior):
    resets=set()
    for b in CA.behavior:
        pos_resets=set()
        for v in b.sensitive:
            if v.isedge():
                resets|=pos_resets
                break
            else:
                pos_resets.add(v.v)
    return resets

# add input port behavior
def add_input_port_behavior(CA, exclude_list={}):
    # add input port behavior
    unspecified_vars = list(CA.state_vars-set(CA.var_spec.keys()))
    unspecified_vars = [var for var in unspecified_vars \
        if var.role == 'input']
    unspecified_vars.sort(lambda x,y: x.name<y.name)
    exclude_vars=exclude_list.keys()
    for v in unspecified_vars:
        if v in exclude_vars:
            guard=['and']
            for ev in exclude_list[v]:
                guard.append(VariableUse(ev, 'stable'))
            guard1=evaluate(guard, {})
            guard2=copy_expression(guard1)
            guard3=evaluate(['not',copy_expression(guard1)], {})
            # (and,c1@stable,c2@stable,...) -> v := v
            b1=Behavior(guard1, v, VariableUse(v, 'last'))
            b1.sensitive=[]
            CA.behavior.add(b1)
            # (and,c1@stable,c2@stable,...) -> v := (not,v)
            b2=Behavior(guard2, v, ['not',VariableUse(v, 'last')])
            b2.sensitive=[]
            CA.behavior.add(b2)
            # (not,(and,c1@stable,c2@stable,...) -> v := v
            b3=Behavior(guard3, v, VariableUse(v, 'last'))
            b3.sensitive=[]
            CA.behavior.add(b3)
            if OPTIONS['verbose']>=3:
                print 'Adding behavior of %s' % v.name
                print '  %s' % colorbehavior(b1)
                print '  %s' % colorbehavior(b2)
                print '  %s' % colorbehavior(b3)
        else:
            b1=Behavior(1, v, VariableUse(v, 'last'))
            b1.sensitive=[]
            CA.behavior.add(b1)
            b2=Behavior(1, v, ['not', VariableUse(v, 'last')])
            b2.sensitive=[]
            CA.behavior.add(b2)
            if OPTIONS['verbose']>=3:
                print 'Adding behavior of %s' % v.name
                print '  %s' % colorbehavior(b1)
                print '  %s' % colorbehavior(b2)

def add_constants_to_behavior(variables, behavior, constants):
    for b in behavior: # a set of Behavior(guard,dest,value)
        b.guard = evaluate(b.guard, {})
    for c,value in constants.iteritems():
        try:
            c=[v for v in variables if v.name==c][0]
        except:
            t_error('Unknown variable %s' % c)
        c.flag='nonstate'
        b=Behavior(1, c, value)
        b.sensitive=[]
        behavior.add(b)

# get variable flags for saving it and later restoring
def get_variable_flags(variables):
    return [(v,v.flag) for v in variables]

# restore variable flags
def restore_variable_flags(flags):
    for var,oldflag in flags:
        var.flag = oldflag

# input: variables, behavior, initlist
# output: outgoing init transition with attributes:
#       s1=new init state (id==-1)
#       guard=1
#       actions=init actions over counters
def make_init_transition(variables, behavior, initlist):
    def find_var(varname):
        vars = [var for var in variables if var.name == varname]
        if vars:
            return vars[0]
        else:
            return None
    def raise_undefined(undefined):
        if len(undefined)>1:
            delimiter='\n    '
        else:
            delimiter=' '
        undefined = ['%-5s %s' % (find_var(var).type, var) \
            for var in undefined]
        undefined.sort()
        t_error('unable to obtain initial value of:%s%s' % \
            (delimiter, delimiter.join(undefined)))
        
    # create a copy of behavior
    behavior_copy=set()
    for item in behavior:
        behavior_copy.add(item.copy())
    behavior=behavior_copy
    del behavior_copy
    # save old flags
    flags = get_variable_flags(variables)
    # mark variables by their types
    mark_state_variables(variables, behavior)
    mark_nonstate_variables(variables, behavior)
    mark_counters(variables, behavior)
    sv = set([v.name for v in variables if v.flag.startswith('state')])
    counters = set([v.name for v in variables if v.flag=='counter'])
    add_constants_to_behavior(variables, behavior, initlist)
    isv = set([v.name for v in variables if v.flag.startswith('state')])
    substitute_nonstate_variables(variables, behavior)
    # restore old variable flags
    restore_variable_flags(flags)
    
    # reduce invalid behavior
    invalid=set()
    for b in behavior:
        b.guard=evaluate(b.guard, {})
        b.value=evaluate(b.value, {})
        if not b.guard:
            invalid.add(b)
    behavior -= invalid
    # create map: variable name->set of its behavior
    var_spec={}
    for b in behavior:
        if b.var.name in var_spec.keys():
            var_spec[b.var.name].append(b)
        else:
            var_spec[b.var.name]=[b]
    # every isv must have one behavior with 1 guard and one of {0,1} value
    # every counter must have one behavior with 1 guard
    # undefined: list of undefined isv or counter
    var_spec={}
    for b in behavior:
        if b.var.name in var_spec.keys():
            var_spec[b.var.name].append(b)
        else:
            var_spec[b.var.name]=[b]
    undefined=(isv|counters)-set(var_spec.keys())
    # unspecified: check for one behavior, guards=1, and {0,1} isv value
    unspecified=set()
    specified={}
    actions=[]
    for var,beh in var_spec.iteritems():
        # one behavior only
        if len(beh)!=1:
            unspecified.add(var)
        elif beh[0].guard!=1:
            unspecified.add(var)
        elif (var in isv) and (beh[0].value not in [0,1]):
            unspecified.add(var)
        elif var in isv:
            specified[var]=beh[0].value
        elif var in counters:
            actions.append(Action(beh[0].var, beh[0].value))
        else:
            assert 0
    undefined |= (sv-isv)-set(initlist.keys())
    undefined |= unspecified
    if undefined:
        raise_undefined(undefined)
    # init state evaluation
    s=initlist.copy()
    for key,value in specified.iteritems():
        s[key]=value
    init_trans=Transition(State(s, 'init'), None, 1, None)
    init_trans.actions=actions
    return init_trans

def connect_init_state(ca):
    """connect_init_state(ca) - connects init state to appropriate state
    in current set of states (with the same state variable evaluation)"""
    for s in ca.states:
        equals=True
        for var,value in s.v.iteritems():
            if ca.init_state.v[var]!=value:
                equals=False
                break
        if equals:
            ca.init_transition.s2 = s
            break
    if not equals:
        t_error('unable to find target of init transition from %s' % \
            ca.init_state)
    else:
        ca.transitions.append(ca.init_transition)

def make_bad_state(variables, behavior):
    """make_bad_state(variables, bahavior)
       returns State from OPTIONS['badvar']"""
    badvar_name = OPTIONS['badvar']
    if badvar_name:
        # find bad variable
        badvar = [v for v in variables if v.name == badvar_name]
        if not badvar:
            t_error('Unknown variable %s' % badvar_name)
        badvar = badvar[0]
        # set its role
        badvar.role = 'bad'
        return (badvar, State({badvar_name: 1}, 'bad'))
    else: # not defined bad variable
        return (None, State({}, 'bad'))

def reduce_bad_variable_evaluation(ca):
    """reduce_bad_variable_evaluation(ca) - clear bad variable evaluations
    from every state"""
    badvar_name = OPTIONS['badvar']
    for s in ca.states:
        if s.v.has_key(badvar_name):
            del s.v[badvar_name]

def normal_varuse(expr):
    """normal_varuse(expr) - returns copy of expr with VariableUse@normal
    instead of Variable"""
    if type(expr) == Variable:
        return VariableUse(expr, 'normal')
    elif type(expr) == list:
        return [expr[0]] + [normal_varuse(e) for e in expr[1:]]
    else: # int?
        return expr


###############################################################################
# Main functions - main, help, usage
###############################################################################

# What do you think this function does?
def make_counter_automaton(CA, variables, behavior, exclude={}, \
                           constants={}, initlist={}):
    # Variable.flag:
    #       'unknown' - unknown; will be one of next flags (set by vhdltrans)
    #       'parameter' - variable representing parameter
    #       'counter' - variable is counter
    #       'state' - state variable
    #       'state clock' - state variable (and it is clock) (set by vhdltrans)
    #       'nonstate' - nonstate variable
    # Variable.role:
    #       'input' - input one bit port
    #       'isv' - internal state variable (one bit)
    #       'counter' - counter
    CA.vendor = sys.argv[0].split('/')[-1]
    CA.output_options = OPTIONS['output options']
    CA.init_transition = make_init_transition(variables, behavior, initlist)
    # make some action guards to be 1 or anything else;
    # this is useful for finding state and nonstate variables/counters
    add_constants_to_behavior(variables, behavior, constants)
    mark_state_variables(variables, behavior)
    mark_nonstate_variables(variables, behavior)
    CA.temp_vars=set([v for v in variables if v.flag=='nonstate'])
    CA.temp_vars={}
    for v in [v for v in variables if v.flag=='nonstate']:
        CA.temp_vars[v]=[strexpr(b.value) for b in behavior if b.var==v][0]
    mark_counters(variables, behavior)
    mark_parameters(variables, behavior)
    # make the bad state
    CA.badvar, CA.bad_state = make_bad_state(variables, behavior)
    # change behavior not to use nonstate variables
    substitute_nonstate_variables(variables, behavior)
    # remove nonstate badvar behavior and set its value to badvar.value
    if CA.badvar and CA.badvar.flag == 'nonstate':
        badbeh = [b for b in behavior if b.var == CA.badvar]
        CA.badvar.value = normal_varuse(badbeh[0].value)
        behavior -= set(badbeh)
    # partition variables
    CA.state_vars=set([v for v in variables \
        if v.flag in ['state','state clock']])
    for v in CA.state_vars:
        if type(v.vhdl)==parser.Port and v.vhdl.direction in ('in','inout'):
            v.role='input'
        elif v.role != 'bad':
            v.role = 'isv'
    # prepare sets of variables
    CA.clocks=[v for v in CA.state_vars if v.flag=='state clock']
    CA.counters=set([v for v in variables if v.flag=='counter'])
    for v in CA.counters:
        v.role='counter'
    CA.variables=CA.state_vars | CA.counters
    CA.parameters=set([v for v in variables if v.flag=='parameter'])
    # shorten guards to faster evaluate behaviors
    CA.behavior=cleared_behavior(behavior)

    # some variable values are not defined when no event is triggered,
    # add behavior which preserves variable value
    add_same_value_behavior(CA)
    # generate exclude list from command line arguments
    exclude_list={}
    for var,vlist in exclude.iteritems():
        try: var=[v for v in CA.variables if v.name==var][0]
        except:
            t_error('State variable name expected in exclude list: %s' % var)
        varlist=[]
        for vname in vlist:
            try: varlist.append([v for v in CA.variables if v.name==vname][0])
            except:
                t_error('State variable name expected in exclude list: %s' % \
                    vname)
        exclude_list[var]=varlist
    # find reset signals (to create exclude list automatically if not defined)
    resets=[]
    specified_vars=CA.var_spec.keys()
    for reset in find_resets(CA.behavior):
        if reset not in specified_vars:
            resets.append(reset)
    for reset in resets:
        if reset not in exclude_list.keys():
            exclude_list[reset]=CA.clocks
    # add definition of variable values which still has to be specified
    # e.g. environment behavior (input ports etc.)
    add_input_port_behavior(CA, exclude_list)

    update_var_spec(CA)
    update_used_vars(CA)

    if OPTIONS['verbose']>=3:
        for k,v in CA.var_spec.iteritems():
            print '%s %s' % (colorstr('Behavior of', 'bold'),
                colorstr(k.name, 'red'))
            for b in v:
                print '  %s' % colorbehavior(b)
                print '    sensitivity list: %s' % \
                    ','.join([str(s) for s in b.sensitive])

    if OPTIONS['generate']:
        # generate states
        CA.generate_states()
        # generate transitions
        CA.generate_transitions()
        # add the init state and transition
        CA.init_state = CA.init_transition.s1
        connect_init_state(CA)
        CA.states.append(CA.init_state)
        # add the bad state to set of states
        CA.states.append(CA.bad_state)
        reduce_bad_variable_evaluation(CA)

# prints statistics about automaton
def print_statistics(ca):
    def str_list(l, c='brown', delim='\n    '):
        lst = [colorstr(i.name, c) for i in l]
        lst.sort()
        return delim.join(lst)
   
    if OPTIONS['verbose']>0:
        print '-'*78
        print 'System %s statistics' % colorstr(ca.name, 'bold')
        print '-'*78
        
        print 'State variable(s): %s' % colorstr(len(ca.state_vars), 'red')
        iv=[v for v in ca.state_vars if v.role=='input']
        clocks=[v for v in ca.state_vars if v.flag=='state clock']
        isv=[v for v in ca.state_vars if v.role=='isv']
        print '  input variable(s): %s' % colorstr(len(iv), 'red')
        if iv:
            print '    %s' % str_list(iv)
        print '  internal state variable(s): %s' % colorstr(len(isv), 'red')
        if isv:
            print '    %s' % str_list(isv)
        print '  detected clock(s): %s' % str_list(clocks, 'default', ',')
        
        print 'Counter(s): %s' % colorstr(len(ca.counters), 'red')
        if ca.counters:
            print '    %s' % str_list(ca.counters)
        
        print 'Parameter(s): %s' % colorstr(len(ca.parameters), 'red')
        if ca.parameters:
            print '    %s' % str_list(ca.parameters)

        print 'Number of states: %s' % \
            (colorstr(len(ca.states), 'red'))

        print 'Number of transitions: %s' % \
            (colorstr(len(ca.transitions), 'red'))

        print 'Transformation time: %s' % \
            (colorstr('%s s' % ca.time, 'red'))

    # print synchronization warnings
    if OPTIONS['verbose']>0 and len(ca.syn_problems):
        print 'Synchronization warnings: %s' % \
            colorstr(len(ca.syn_problems), 'red')
        for vname1,vname2,src in ca.syn_problems:
            print '%s: At: %s' % (src[0], src[2])
            print '%s:%d: warning: asynchronous reading of `%s\'' \
                % (src[0], src[1], colorstr(vname1, 'brown'))

    if OPTIONS['verbose']>1:
        print 'Nonstate/unused variables: %s' % \
            colorstr(len(ca.temp_vars.keys()), 'red')
        l=[[i.name,i] for i in ca.temp_vars.keys()]
        l.sort()
        for i in l:
            print '    %s = %s' % (colorstr(i[0], 'brown'),
                colorstr(ca.temp_vars[i[1]], 'blue'))

###############################################################################
def list_output_types():
    import os
    types=[f[len('output_'):-3] for f in os.listdir(APPPATH) if \
                f.startswith('output_') and f.endswith('.py')]
    types.sort()
    print 'Available output types: %s' % ' '.join(types)

_head_msg="Hardware description to counter automaton transformation"
_about_msg="""Author: Ales Smrcka <smrcka@fit.vutbr.cz>, 2007.
License: GNU GPL (http://www.gnu.org/licenses/gpl.html)"""
_usage_msg="syntax: %s [options] file(s)" % sys.argv[0].split('/')[-1]
_options_msg="""Options:
    -b variable mark variable as bad (go to bad state if equals to one)
    -c v=value  Make a constant from the variable (comma separated list of
                evaluation e.g., -c reset=1,wr=0,oe=1)
    -x v:v1,v2  Create exclude list for variable v. List is comma separated
                list of variable names
    -i v=value  Initial transition evaluation over state variables (comma
                separated list of evaluation)
    -t type     Output type of automaton (see -l for list of available types)
    -l          List available automaton output types and exit
    -o file     Counter automaton will be saved to file
    --noca      Do not generate counter automaton (try with -v3)
    -m size     Set maximum resident memory usage (in megabytes)
    -p option1=value1,option2=value2
                Output plugin options
    -v level    Verbosity level
        0       be quiet (except errors and warnings)
        1       progress and statistics (default)
        2       level 1 + nonstate/unused variables
        3       intermediate model
        4+      more debugging"""

def help():
    print _head_msg
    print _about_msg
    print ''
    print _usage_msg
    print _options_msg

def usage():
    print _head_msg
    print _usage_msg

def main(filename_list):
    import time
    start_time = time.time()
    # name is a basename of the first file without an extension
    CA.name='.'.join(filename_list[0].split('/')[-1].split('.')[:-1])
    CA.sources=filename_list
    try:
        output_module=None
        if OPTIONS['output type']:
            try:
                output_module=__import__('output_%s' % OPTIONS['output type'])
            except:
                print >>sys.stderr, \
                    'Something wrong is with output_%s module\n' % \
                    OPTIONS['output type']
                sys.exit(2)
        else:
            output_module=None
        # parse and create syntax structure
        for file in filename_list:
            parser.parse_abstract_file(file)
        (variables, behavior) = read_useful_behavior(parser.entities[0])
        # make counter automaton
        make_counter_automaton(CA, variables, behavior, \
            OPTIONS['excludes'], OPTIONS['constants'], OPTIONS['inits'])
        CA.time = int(time.time() - start_time)
        # output it
        if output_module:
            if OPTIONS['output']:
                o=open(OPTIONS['output'],'w')
            else:
                o=open('%s.%s' % (CA.name,OPTIONS['output type']), 'w')
            o.write(output_module.transformate_automaton(CA))
            o.close()
        print_statistics(CA)
    except parser.CAException, e:
        e.print_error()

def print_C_errorfmt(e):
    "print C error format message; only for debugging in vim"
    import traceback
    l=traceback.extract_tb(sys.exc_traceback)[1:]
    l.reverse()
    for i in l:
        file,line = i[:2]
        if i==l[0]:
            print '%s:%d:%s' % (file, line, str(e))
        else:
            print '%s:%d:%s' % (file, line, i[2])

def memory_limit(sizeM):
    "set limit of max resident memosy usage (in megabytes)"
    import resource
    mb=2**20
    resource.setrlimit(resource.RLIMIT_AS, (sizeM*mb,sizeM*mb))
    del resource

def process_arguments():
    "process sys.argv and set global options"
    import os, getopt
    OPTIONS['color'] = OPTIONS['color'] & os.isatty(1)
    try:
        optlist, filename_list = \
            getopt.getopt(sys.argv[1:], 'b:hlc:i:m:o:p:t:v:x:', ['noca'])
    except getopt.GetoptError:
        usage()
        sys.exit(2)
    try:
        for opt, arg in optlist:
            if opt == '--noca':
                OPTIONS['generate'] = 0
            elif opt == '-b':
                OPTIONS['badvar'] = arg
            elif opt == '-c':
                clist = arg.split(',')
                for constant in clist:
                    var, val = constant.split('=')
                    OPTIONS['constants'][var] = int(val)
            elif opt == '-h':
                help()
                return 0
#                sys.exit(0)
            elif opt == '-i':
                clist = arg.split(',')
                for constant in clist:
                    var, val = constant.split('=')
                    OPTIONS['inits'][var] = int(val)
            elif opt == '-l':
                list_output_types()
            elif opt == '-o':
                OPTIONS['output'] = arg
            elif opt == '-p':
                olist = arg.split(',')
                for oitem in olist:
                    option, val = oitem.split('=')
                    OPTIONS['output options'][option] = val
            elif opt == '-t':
                OPTIONS['output type'] = arg
            elif opt == '-x':
                var, xlist = arg.split(':')
                OPTIONS['excludes'][var] = xlist.split(',')
            elif opt == '-m':
                memory_limit(int(arg))
            elif opt == '-v':
                OPTIONS['verbose'] = int(arg)
    except:
        print >>sys.stderr, 'invalid argumets'
        print >>sys.stderr, _usage_msg
        sys.exit(2)
    if not filename_list:
        print >>sys.stderr, '%s:no input file specified' % sys.argv[0]
        print >>sys.stderr, 'for help: %s -h' % sys.argv[0]
        print >>sys.stderr, _usage_msg
        sys.exit(2)
    return filename_list

if __name__=="__main__":
    APPPATH = '/'.join(sys.argv[0].split('/')[:-1])
    filename_list = process_arguments()
    if filename_list == 0:
        sys.exit(0)
    try:
        import vhdl_parser as parser
        from simplify import * # simplify.py
        from vhdltrans import *
        main(filename_list)
    except KeyboardInterrupt:
        if current_progress: current_progress.stop()
        print >>sys.stderr, 'User interrupt. No automaton was generated.'
    except TransformationError, e:
        if current_progress: current_progress.stop()
        print >>sys.stderr, 'error:%s' % e
    except MemoryError:
        if current_progress: current_progress.stop()
        print >>sys.stderr, 'error:%s' % \
            'System interrupt. Maximum resident memory exceeded.'
    except Exception, e:
#        print_C_errorfmt(e)
        print >>sys.stderr, 'error:%s' % e
