#! /usr/bin/env python
# vim:ts=4:expandtab:sw=4:tw=0
###############################################################################
# vhdl_parser.py - lexer and parser for VHDL                                  #
#                                                                             #
# $Id: vhdl_parser.py 202 2007-06-10 01:11:52Z smrcka $
#                                                                             #
# Author: Ales Smrcka <smrcka@fit.vutbr.cz>                                   #
#                                                                             #
# This program is free software; you can redistribute it and/or modify        #
#   it under the terms of the GNU General Public License as published by      #
#   the Free Software Foundation; either version 2 of the License, or         #
#   (at your option) any later version.                                       #
#   See http://www.gnu.org/licenses/gpl.html for more details.                #
#                                                                             #
###############################################################################
###############################################################################
# tokens based on liberouter VHDL files
# * grammar based on VHDL Quick Reference Card (95-00,Qualis Design Corp.)
#   http://soc.eurecom.fr/EDC/vhdlref.pdf
# * 1164 pkg quick reference card: http://soc.eurecom.fr/EDC/1164pkg.pdf
# 
#
# Owing to simplicity, VHDL source code is translated before parsing:
# <<<
# -- abstract begin optional comment/identification/anything
# -- something_abstract1
# something_concrete1
# -- something_abstract2
# -- end abstract optional anything
# >>>
# -- abstract begin optional comment/identification/anything
# something_abstract1
# -- something_concrete1
# something_abstract2
# -- something_concrete2
# -- end abstract optional anything
#

###############################################################################
# LEXER
###############################################################################
import lex

# reserved tokens
tokens_reserved = (
    # key words
    'ENTITY', 'ARCHITECTURE', 'PACKAGE', 'CONFIGURATION', 'LIBRARY', 'USE',
    'GENERIC', 'PORT', 'BEGIN', 'END', 'IS', 'FOR', 'OF', 'ALL', 'MAP',
    'TYPE', 'RANGE', 'TO', 'DOWNTO', 'ARRAY', 'RECORD', 'BODY',
    'IN', 'OUT', 'INOUT', 'ALIAS', 'ATTRIBUTE',
    'SIGNAL', 'CONSTANT', 'COMPONENT', 'SUBTYPE', 'ACCESS',
    'VARIABLE', 'OTHERS', 'PROCEDURE', 'FUNCTION', 'LABEL',
    'PROCESS', 'IF', 'THEN', 'ELSIF', 'LOOP', 'GENERATE',
    'RETURN', 'FILE', 'NEW', 'NULL', 'WAIT', 'UNTIL', 'ON',
    'ASSERT', 'WHEN', 'NEXT', 'EXIT', 'REPORT', 'SEVERITY', 'NOTE',
    'WARNING', 'ERROR', 'FAILURE', 'TRANSPORT', 'AFTER',
    'ELSE', 'CASE', 'WHILE', 'BLOCK', 'WITH', 'SELECT',
    # VHDL-1993
    'SHARED', 'OPEN', 'READ_MODE', 'WRITE_MODE', 'APPEND_MODE',
    'IMPURE', 'POSTPONED', 'INERTIAL', 'UNAFFECTED',
    # binary operators
    'AND', 'NAND', 'OR', 'NOR', 'XOR', 'XNOR',
    'MOD', 'REM',
    # unary operators
    'ABS', 'NOT',
    # extension for transformation tool
    'CONSTRAINT',
)

# string values for reserved tokens
RESERVED_TYPES = {}
for token in tokens_reserved:
    RESERVED_TYPES[token.lower()] = token
del token

# tokens of symbols
tokens_symbols = (
    'ASSIGN', 'EXPSIGN', 'APOSTROPHE', 'CONNECT',
    'EQ', 'NEQ', 'LE', 'LT', 'GE', 'GT', 'RANGESIGN',
    # due to bug in ply-2.[0-3], DOT is a special token for '.'
    'DOT',
)

# list of all tokens (for lex)
tokens = tokens_reserved + tokens_symbols + (
    'ID', 'LITERAL', 'INTLITERAL',
)

# one char literals (for lex)
literals = (
    ';', '(', ')', ',', ':',
    '+', '-', '&', '*', '/', '|',
)

# token definitions
t_ignore = " \t"

t_DOT = r'\.'
t_GE = r'>='
t_LE = r'<='
t_EQ = r'='
t_NEQ = r'/='
t_LT = r'<'
t_GT = r'>'
t_ASSIGN = r':='
t_CONNECT = r'=>'
t_EXPSIGN = r'\*\*'
t_APOSTROPHE = r'\''
t_RANGESIGN = r'<>'

# VHDL comment or abstract specification
def t_ignore_COMMENT(t):
    r'--.*\n'
    t.lexer.lineno += 1

# int value -> str in binary
def digit2bin(value, length):
    s=''
    while value:
        s = str(value&1)+s
        value = value>>1
    s = '0'*(length-len(s))+s
    return s

# VHDL LITERAL (INTLITERAL) - integer number or bit vector ('string')
# returns either int or str value
def t_LITERAL(t):
    r'(\d+|[BOX]?\'[0-9a-fA-F]+\'|[BOX]?".*")'
    if t.value[0]=='B':             # binary string (return bin str)
        t.value = t.value[2:-1]
    elif t.value[0]=='O':           # octal string (return bin str)
        s=''
        for c in t.value[2:-1]:
            s = s+digit2bin(int(c), 3)
        t.value = s
    elif t.value[0]=='X':           # hex string (return bin str)
        s=''
        for c in t.value[2:-1]:
            c=c.lower()
            if c.isdigit():
                s = s+digit2bin(int(c), 4)
            else:
                s = s+digit2bin(ord(c)-ord('a')+10, 4)
        t.value = s
    elif t.value[0] in ("'", '"'):           # binary string (return bin str)
        t.value = t.value[1:-1]
    else:
        t.value = int(t.value)      # a number (return int)
        t.type = 'INTLITERAL'
    return t

# VHDL ID - regexp is compatible with key words, 
# token type must be distinguished
def t_ID(t):
    r'[a-zA-Z_][a-zA-Z0-9_]*'
    if t.value.upper() in tokens_reserved:
        t.type = t.value.upper()
    t.value = t.value.lower()
    return t

def t_newline(t):
    r'\n+'
    t.lexer.lineno += t.value.count('\n')

def t_error(t):
    print "%s:%d:illegal character '%s'" % \
        (t.lexer.filename, t.lexer.lineno, t.value[0])

# build the lexer
lexer=lex.lex()
###############################################################################

###############################################################################
# PARSER
###############################################################################
from component import *

###############################################################################
# Functions for syntax structure
###############################################################################

# common rule for generating token tree
def common_rule(p):
    "common_rule(list_of_tokens) returns str(token) or list of tokens"
    def reduce_list(i):
        if type(i)==list:
            if len(i)==0:
                return None
            elif len(i)==1:
                return reduce_list(i[0])
            else:
                return i
        else:
            return i
    result=[]
    for i in p:
        if type(i)==list:
            result.append(reduce_list(i))
        elif i!=None:
            result.append(i)
    return reduce_list(result)

# generate list from list productions
def generate_list(p, item_idx, list_idx):
    if item_idx>=len(p):
        p[0]=[]
    elif list_idx<len(p):
        p[0]=[p[item_idx]]+p[list_idx]
    else:
        p[0]=[p[item_idx]]


# True if we are inside supported section (use when generating Usage objects)
supported=True

# currently generated object(s) (stack of VHDLobject)
syntax_stack=[]

def get_parent(vhdltype=VHDLobject, f=None):
    """get_parent(vhdltype=VHDLobject, f=None) returns last pushed object of
    vhdltype, f is one-arg filter function"""
    c=syntax_stack[:]
    c.reverse()
    for i in c:
        if vhdltype in type(i).mro():
            if f:
                if f(i):
                    return i
            else:
                return i
    return None

def pop_current(vhdltype, f=None):
    """pop_current(vhdltype, f=None) pops all objects up to vhdltype object
    which is returned, f is one-arg filter function"""
    global syntax_stack
    while len(syntax_stack):
        i=syntax_stack.pop()
        if vhdltype in type(i).mro():
            if f:
                if f(i):
                    return i
            else:
                return i
    return None

def push_current(vhdlobject):
    "push_current(vhdlobject) push vhdlobject to syntax stack"
    global syntax_stack
    syntax_stack.append(vhdlobject)

# search for an object in syntax stack which has set the declaring attribute
def find_parent():
    return get_parent(VHDLobject, lambda x: x.declaring)

# from id search for an object (in syntax stack)
def search_id(p, id, vhdltypes=[Signal, Constant, SignalEvent, \
    SignalPosedge, SignalNegedge, Parameter]):
    """seach_id(id, vhdltypes=[Signal, Constant, SignalEvent, Parameter])
    searches for VHDLobject in current namespace, vhdltypes is a list of
    supported VHDLobject classes. Returns VHDLobject or None if not found."""
    g=get_parent(GlobalObject)
    if g:
        found_object=g[id]
        if set(type(found_object).mro()).intersection(vhdltypes):
            return found_object
    VHDLobject(p).vhdl_error('unknown identificator %s' % id)

###############################################################################
# 0. Common Productions
###############################################################################

# start symbol
def p_start(p):
    """start : library start
             | empty"""
    p[0]=common_rule(p[1:]) #pass

# empty rule
def p_empty(p):
    "empty :"
    p[0]=None

###############################################################################

# match: [:= expr]
def p_assign_expr_opt(p):
    """assign_expr_opt : ASSIGN expr
                       | empty"""
    if p[1]:
        p[0]=p[2]
        for s in p[2].get_evals():
            s.usage.append(Usage(p, access=Usage.R))
    else:
        p[0]=None

# match: {range,}
def p_range_list(p):
    """range_list : range
                  | range ',' range_list"""
    generate_list(p, 1, 3)

# match: range
def p_range(p):
    """range : expr range_dir expr
             | '(' ID RANGE RANGESIGN ')'"""
    if len(p)==4:
        for s in p[1].get_evals():
            s.usage.append(Usage(p, access=Usage.R))
        for s in p[3].get_evals():
            s.usage.append(Usage(p, access=Usage.R))
        if p[2]=='to':
            p[0]=Range(p, bottom=p[1], top=p[3])
        else: #downto
            p[0]=Range(p, bottom=p[3], top=p[1])
    else:
        p[0]=c=Range(p)
        c.ignore(common_rule(p[1:]), 'range')

# match: to|downto
def p_range_dir(p):
    """range_dir : TO
                 | DOWNTO"""
    p[0]=p[1]

# match: [generic ( {ID:TYPEID [:= expr];} );]
def p_generic_opt(p):
    """generic_opt : GENERIC '(' id_type_expr_list ')' ';'
                   | empty"""
    if p[1]:
        p[0]=p[3]
    else:
        p[0]=[]

# match: {ID:TYPEID [:= expr];}
def p_id_type_expr_list(p):
    """id_type_expr_list : id_type_expr
                         | id_type_expr ';' id_type_expr_list"""
    generate_list(p, 1, 3)

# match: ID:TYPEID [:= expr]
def p_id_type_expr(p):
    "id_type_expr : ID ':' typeid assign_expr_opt"
    p[0]=c=Parameter(p, id=p[1], type=p[3])
    if len(p)>=5:
        c.value=p[4]

# match: [port ( {ID: in|out|inout TYPEID [:= expr];} );]
def p_port_opt(p):
    """port_opt : PORT '(' id_port_type_expr_list ')' ';'
                | empty"""
    if p[1]:
        p[0]=p[3]
    else:
        p[0]=[]

# match: {ID: in|out|inout TYPEID [:= expr];}
def p_id_port_type_expr_list(p):
    """id_port_type_expr_list : id_port_type_expr
                              | id_port_type_expr ';' id_port_type_expr_list"""
    generate_list(p, 1, 3)

# match: in|out|inout
def p_port_type(p):
    """port_type : IN
                 | OUT
                 | INOUT"""
    p[0]=p[1].lower()

# match: ID: in|out|inout TYPEID [:= expr]
def p_id_port_type_expr(p):
    "id_port_type_expr : ID ':' port_type typeid assign_expr_opt"
    p[0]=c=Port(p, id=p[1], direction=p[3], type=p[4], value=p[5])
    if c.value:
        c.usage.append(Usage(p, access=Usage.W))

# match: {[constant|variable|signal] ID : in|out|inout TYPEID [:= expr];}
def p_parameter_expr_list(p):
    """parameter_expr_list : parameter_expr
                           | parameter_expr ';' parameter_expr_list"""

# match: [constant|variable|signal] ID : in|out|inout TYPEID [:= expr]
def p_parameter_expr(p):
    """parameter_expr : CONSTANT ID ':' port_type typeid assign_expr_opt
                      | VARIABLE ID ':' port_type typeid assign_expr_opt
                      | SIGNAL ID ':' port_type typeid assign_expr_opt
                      | ID ':' port_type typeid assign_expr_opt"""
    p[0]=common_rule(p[1:]) #pass

# match: {ID:TYPEID;}
def p_id_type_list(p):
    """id_type_list : id_type
                    | id_type ';' id_type_list"""
    p[0]=common_rule(p[1:]) #pass

# match: ID:TYPEID
def p_id_type(p):
    "id_type : ID ':' typeid"
    p[0]=common_rule(p[1:]) #pass

# match: {ID: in|out|inout TYPEID;}
def p_id_port_type_list(p):
    """id_port_type_list : id_port_type
                         | id_port_type ';' id_port_type_list"""
    p[0]=common_rule(p[1:]) #pass

# match: ID: in|out|inout TYPEID
def p_id_port_type(p):
    "id_port_type : ID ':' port_type typeid"
    p[0]=common_rule(p[1:]) #pass

# match: [[generic map ( {ID => expr,} )] port map ( {ID => ID|expr ,} )]
def p_map_opt(p):
    """map_opt : generic_map_opt port_map
               | empty"""
    p[0]=common_rule(p[1:]) #pass

# match: [generic map ( {ID => expr,} )]
def p_generic_map_opt(p):
    """generic_map_opt : GENERIC MAP '(' id_map_list ')'
                       | empty"""
    p[0]=common_rule(p[1:]) #pass

# match: {ID => expr,}
def p_id_map_list(p):
    """id_map_list : id_map
                   | id_map ',' id_map_list"""
    p[0]=common_rule(p[1:]) #pass

#match: ID => expr
def p_id_map(p):
    """id_map : ID CONNECT expr
              | ID CONNECT OPEN"""
    p[0]=common_rule(p[1:]) #pass

# match: port map ( {[ID =>] expr ,} )
def p_port_map(p):
    "port_map : PORT MAP '(' id_map_list ')'"
    p[0]=common_rule(p[1:]) #pass

# match: [generic map ( {[ID =>] expr,} )]
def p_generic_map_key_opt(p):
    """generic_map_key_opt : GENERIC MAP '(' id_map_key_opt_list ')'
                           | empty"""
    p[0]=common_rule(p[1:]) #pass

# match: {[ID =>] expr,}
def p_id_map_key_opt_list(p):
    """id_map_key_opt_list : id_map_key_opt
                           | id_map_key_opt ',' id_map_key_opt_list"""
    p[0]=common_rule(p[1:]) #pass

#match: [ID =>] expr
def p_id_map_key_opt(p):
    """id_map_key_opt : expr
                      | ID CONNECT expr
                      | ID CONNECT OPEN"""
    p[0]=common_rule(p[1:]) #pass

# match: port map ( {[ID =>] expr ,} )
def p_port_map_key_opt(p):
    "port_map_key_opt : PORT MAP '(' id_map_key_opt_list ')'"
    p[0]=common_rule(p[1:]) #pass

# match: TYPEID
def p_typeid1(p):
    "typeid : ID"
    if p[1]=='integer':
        p[0]=Int
    elif p[1]=='std_logic':
        p[0]=Bool
    else:
        p[0]=c=Type(p)
        c.vhdl_error('unsupported type')

def p_typeid__range(p):
    "typeid : ID '(' range ')'"
    if p[1]=='std_logic_vector':
        p[0]=BinVectorType(p, range=p[3])
    else:
        p[0]=c=Type(p)
        c.vhdl_error('unsupported type')

def p_typeid__parid_list(p):
    "typeid : ID '(' parid_list ')'"
    p[0]=c=Type(p)
    c.vhdl_error('unsupported type')

# match: [is]
def p_is_opt(p):
    """is_opt : IS
              | empty"""
    p[0]=common_rule(p[1:]) #pass

# match: [ID]
def p_id_opt(p):
    """id_opt : ID
              | empty"""
    p[0]=common_rule(p[1:]) #pass

# match: {expr [after time],}
def p_expr_list1(p):
    "expr_list : expr expr_list_after_opt"
    # TODO usage
    p[0]=p[1]
def p_expr_list2(p):
    "expr_list : expr expr_list_after_opt ',' expr_list"
    # TODO usage
    p[0]=c=VHDLobject(p)
    c.vhdl_error('unsupported expression list')

# match: [after time]
def p_expr_list_after_opt(p):
    """expr_list_after_opt : AFTER expr
                           | empty"""
    # TODO usage
    if p[1]:
        p[0]=c=VHDLobject(p)
        c.ignore(common_rule(p[1:]))
    else:
        p[0]=None

# match: INTLITERAL ID (where ID must be one of hr|min|sec|ms|us|ns|ps|fs)
def p_time(p):
    "time : INTLITERAL ID"
    if p[2].lower() not in ['hr','min','sec','ms','us','ns','ps','fs']:
        raise SyntaxError, 'Wrong time unit %s' % p[2]
    p[0]=Time(p, value=p[1], unit=p[2])

# match: {choice |}
def p_choice_list(p):
    """choice_list : choice
                   | choice '|' choice_list"""
    if len(p)>2:
        p[0]=c=VHDLobject(p)
        c.ignore(common_rule(p[2:]), 'choice_list')
    p[0]=p[1]

# match: choice
def p_choice(p):
    """choice : range
              | OTHERS
              | sexpr"""
    if type(p[1])==Range(): # range
        p[0]=c=VHDLobject(p)
        c.ignore(p[1], 'choice')
    elif Expr in type(p[1]).mro(): # sexpr
        p[0]=c=VHDLobject(p)
        c.ignore(p[1], 'choice')
    else: # 'others'
        p[0]=p[1]

###############################################################################
# 1. Library Units
###############################################################################

#-------- library
def p_library(p):
    """library : use_clause
               | entity ';'
               | architecture ';'
               | package ';'
               | package_body ';'
               | configuration ';'"""
    p[0]=common_rule(p[1:])

#-------- use clause
def p_use_clause(p):
    """use_clause : LIBRARY ID ';'
                  | LIBRARY ID ';' use_list"""
    p[0]=c=Library(p, id=p[2])
    if len(p)>=5:
        c.packages=p[4]

def p_use_list(p):
    """use_list : use
                | use use_list"""
    if len(p)>=3:
        p[0]=[p[1]]+p[2]
    else:
        p[0]=[p[1]]

def p_use(p):
    """use : USE ID DOT ID ';'
           | USE ID DOT ID DOT ID ';'
           | USE ID DOT ID DOT ALL ';'"""
    p[0]=c=Package(p)
    if len(p)>=7:
        c.use='%s.%s.%s' % (p[2],p[4],p[6])
    else:
        c.use='%s.%s' % (p[2],p[4])

#-------- entity unit
def p_entity(p):
    "entity : entity_header generic_opt entity1 constraint_list_opt port_opt declaration_list entity_body_opt entity_tail"
    c=p[1]
    #c.ports=p[5]
    c.register(p[5])
    if p[6]: c.ignore(p[6], 'declaration_list')
    if p[7]: c.ignore(p[7], 'entity_body_opt')
    c.declaring=False
    pop_current(Entity)
    # global entities
    entities.append(c)

def p_entity_header(p):
    "entity_header : ENTITY ID IS"
    p[0]=c=Entity(p, id=p[2])
    c.declaring=True
    push_current(c)

# transformation tool extension to VHDL source
def p_constraint_list_opt(p):
    """constraint_list_opt : constraint_list
                           | empty"""
    pass

def p_constraint_list1(p):
    "constraint_list : constraint ';'"
    pass
def p_constraint_list2(p):
    "constraint_list : constraint ';' constraint_list"
    pass

def p_constraint(p):
    "constraint : CONSTRAINT expr"
    # example of use:
    # generic (ITEMS: integer);
    # constraint ITEMS>1;
    # port (...);
    expr = p[2]
    if Prim in type(expr).mro():
        if expr.prim != 1:
            VHDLobject(p).vhdl_error('invalid constraint (did you mean 1?)')
    elif OpExpr in type(expr).mro():
        try:
            parameter = expr.operands[0].prim
            assert Parameter in type(parameter).mro()
            parameter.constraint = expr
        except:
            VHDLobject(p).vhdl_error('invalid constraint')
    else:
        VHDLobject(p).vhdl_error('invalid constraint')

def p_entity1(p):
    "entity1 :"
    # we must register parameter declaration before any other declaration
    c=p[-2]
    #c.parameters=p[-1]
    c.register(p[-1])

def p_entity_body_opt(p):
    """entity_body_opt : BEGIN parallel_stmt_list END
                       | END"""
    p[0]=common_rule(p[1:]) #pass

def p_entity_tail(p):
    """entity_tail : ID
                   | ENTITY ID
                   | ENTITY
                   | empty"""
    pass

#-------- architecture unit
def p_architecture(p):
    "architecture : architecture_header declaration_list architecture1 BEGIN parallel_stmt_list_opt architecture_tail"
    # parse step: 4 (final)
    c=p[1]
    for i in p[5]:
        if type(i)==CondAssign:
            c.connections.append(i)
        elif type(i)==Process:
            c.processes.append(i)
            c.connections+=i.assigns
        else:
            c.ignore(i, 'parallel_stmt')
    c.declaring=False
    pop_current(Architecture)

def p_architecture_header(p):
    "architecture_header : ARCHITECTURE ID OF ID IS"
    # parse step: 1
    p[0]=c=Architecture(p, id=p[2])
    # we start to declare here
    c.declaring=True
    try:
        entity=id_map[p[4]]
        if type(entity)==Entity:
            c.parent=entity
            entity.architecture=c
        else:
            c.vhdl_error('%s is not an entity identificator' % p[4])
    except:
        c.vhdl_error('unknown entity %s' % p[4])
    push_current(c) # push Architecture to generated objects

def p_architecture1(p):
    "architecture1 :"
    # parse step: 2

def p_architecture_tail(p):
    """architecture_tail : END ARCHITECTURE ID
                         | END ARCHITECTURE
                         | END ID
                         | END"""
    # parse step: 3
    pass

#-------- package unit
def p_package(p):
    "package : package_header declaration_list package_tail"
    p[0]=c=p[1]
    c.ignore(p[2], 'declaration_list')

def p_package_header(p):
    "package_header : PACKAGE ID IS"
    p[0]=PackageUnit(p, id=p[2])

def p_package_tail(p):
    """package_tail : END PACKAGE ID
                    | END ID"""
    pass

#-------- package body unit
def p_package_body(p):
    "package_body : package_body_header declaration_list package_body_tail"
    p[0]=c=p[1]
    c.ignore(p[2], 'declaration_list')

def p_package_body_header(p):
    "package_body_header : PACKAGE BODY ID IS"
    p[0]=PackageBody(p, id=p[3])

def p_package_body_tail(p):
    """package_body_tail : END PACKAGE BODY ID
                         | END ID"""
    pass

#-------- configuration
def p_configuration(p):
    "configuration : configuration_header FOR ID config_list configuration_tail"
    p[0]=c=p[1]
    c.ignore(p[4], 'config_list: ARCHID=%s' % p[3])

def p_configuration_header(p):
    "configuration_header : CONFIGURATION ID OF ID IS"
    p[0]=Configuration(p, id=p[2], entity=p[4])

def p_configuration_tail(p):
    """configuration_tail : END FOR ';' CONFIGURATION ID
                          | END FOR ';' ID"""
    pass

#---- config list
# match: [{block_config|comp_config}]
def p_config_list(p):
    """config_list : block_config config_list
                   | comp_config config_list
                   | empty"""
    if p[1]:
        generate_list(p, 1, 2)
    else:
        p[0]=[]

#---- block config
def p_block_config(p):
    "block_config : FOR ID config_list END FOR ';'"
    p[0]=c=BlockConfig(p, label=p[2])
    c.ignore(p[3], 'config_list')

#---- comp config
def p_comp_config(p):
    """comp_config : FOR ALL ':' ID comp_entity END FOR ';'
                   | FOR ALL ':' ID comp_configuration END FOR ';'
                   | FOR ID ':' ID comp_entity END FOR ';'
                   | FOR ID ':' ID comp_configuration END FOR ';'"""
    p[0]=c=ConfConfig(p)
    c.ignore(p[5], 'comp_entity | comp_configuration')

def p_comp_entity(p):
    """comp_entity : USE ENTITY ID DOT ID '(' ID ')' map_opt ';' comp_entity_for_opt END FOR ';'
                   | USE ENTITY ID DOT ID map_opt ';' comp_entity_for_opt END FOR ';'
                   | USE ENTITY ID '(' ID ')' map_opt ';' comp_entity_for_opt END FOR ';'
                   | USE ENTITY ID map_opt ';' comp_entity_for_opt END FOR ';'"""
    p[0]=common_rule(p[1:]) #pass

def p_comp_entity_for_opt(p):
    """comp_entity_for_opt : FOR ID config_list END FOR ';'
                           | empty"""
    p[0]=common_rule(p[1:]) #pass

def p_comp_configuration(p):
    """comp_configuration : USE CONFIGURATION ID DOT ID map_opt ';'
                          | USE CONFIGURATION ID map_opt ';'"""
    p[0]=common_rule(p[1:]) #pass

###############################################################################
# 2. Declaration
###############################################################################

#-------- declaration list
# match: [{declaration}]
def p_declaration_list(p):
    """declaration_list : declaration ';' declaration_list
                        | empty"""
    if p[1]:
        generate_list(p, 1, 3)
    else:
        p[0]=[]

def p_declaration(p):
    """declaration : decl_type
                   | decl_constant
                   | decl_variable
                   | decl_signal
                   | decl_file
                   | decl_alias
                   | decl_attrib
                   | decl_component
                   | decl_function
                   | decl_procedure
                   | decl_for"""
    p[0]=c=p[1]
    c.parent=find_parent() # set parent and register this object

#-------- Type Declarations
def p_decl_type(p):
    """decl_type : TYPE ID IS decl_type_def
                 | SUBTYPE ID IS decl_type_subdef"""
    if p[1]=='type':
        p[0]=c=Type(p, id=p[2])
        c.ignore(p[4], 'decl_type_def')
    else:
        p[0]=c=SubType(p, id=p[2])
        c.ignore(p[4], 'decl_type_subdef')

#---- decl_type TYPE definition
def p_decl_type_def(p):
    """decl_type_def : '(' id_list ')'
                     | RANGE INTLITERAL range_dir INTLITERAL
                     | ARRAY '(' decl_type_range_list ')' OF expr
                     | RECORD decl_type_record_item_list END RECORD
                     | ACCESS typeid
                     | FILE OF typeid"""
    p[0]=common_rule(p[1:]) #pass

def p_id_list1(p):
    "id_list : ID"
    p[0]=[search_id(p, p[1], [Signal, Constant, Parameter])]

def p_id_list2(p):
    "id_list : ID ',' id_list"
    p[0]=[search_id(p, p[1], [Signal, Constant, Parameter])]+p[3]

def p_decl_type_range_list(p):
    """decl_type_range_list : range
                            | typeid
                            | range ',' decl_type_range_list
                            | typeid ',' decl_type_range_list"""
    generate_list(p, 1, 3)

def p_decl_type_record_item_list(p):
    """decl_type_record_item_list : ID ':' typeid ';'
                                  | ID ':' typeid ';' decl_type_record_item_list"""
    p[0]=common_rule(p[1:]) #pass

#---- decl_type SUBTYPE definition
def p_decl_type_subdef(p):
    """decl_type_subdef : ID RANGE range ';'
                        | ID typeid ';'
                        | ID '(' range_list ')'"""
    p[0]=common_rule(p[1:]) #pass

#-------- Other Declarations
#---- constant declaration
def p_decl_constant(p):
    "decl_constant : CONSTANT ID ':' typeid ASSIGN expr"
    p[0]=c=Constant(p, id=p[2], type=p[4], value=p[6])
    c.usage.append(Usage(p, access=Usage.W))
    for e in p[6].get_evals():
        e.usage.append(Usage(p, access=Usage.R))

#---- variable declaration
def p_decl_variable(p):
    """decl_variable : SHARED VARIABLE ID ':' typeid assign_expr_opt
                     | VARIABLE ID ':' typeid assign_expr_opt"""
    shared=0
    if p[1]=='shared': shared=1
    p[0]=c=Variable(p, id=p[shared+2])
    c.ignore(p[shared+4:shared+5], 'typeid, assign_expr_opt')

#---- signal declaration
def p_decl_signal(p):
    """decl_signal : SIGNAL ID ':' typeid assign_expr_opt"""
    p[0]=c=Signal(p, id=p[2], type=p[4])
    if p[5]:
        c.value=p[5]
        c.usage.append(Usage(p, access=Usage.W))

#---- file declaration
def p_decl_file(p):
    """decl_file : FILE ID ':' typeid IS inout LITERAL
                 | FILE ID ':' typeid IS OPEN file_open_mode IS LITERAL"""
    p[0]=File(p, id=p[2])

def p_inout(p):
    """inout : IN
             | OUT"""
    p[0]=common_rule(p[1:]) #pass

def p_file_open_mode(p):
    """file_open_mode : READ_MODE
                      | WRITE_MODE
                      | APPEND_MODE"""
    p[0]=common_rule(p[1:]) #pass

#---- alias declaration
def p_decl_alias(p):
    "decl_alias : ALIAS ID ':' typeid IS ID ';'"
    p[0]=Alias(p, id=p[2])

#---- attribute declaration
def p_decl_attrib(p):
    """decl_attrib : ATTRIBUTE ID ':' typeid ';'
                   | ATTRIBUTE ID OF decl_attrib_which ':' decl_class IS expr ';'"""
    p[0]=Attribute(p, id=p[2])

def p_decl_attrib_which(p):
    """decl_attrib_which : ID
                         | OTHERS
                         | ALL"""
    p[0]=common_rule(p[1:]) #pass

def p_decl_class(p):
    """decl_class : ENTITY
                  | ARCHITECTURE
                  | CONFIGURATION
                  | PROCEDURE
                  | FUNCTION
                  | PACKAGE
                  | TYPE
                  | SUBTYPE
                  | CONSTANT
                  | SIGNAL
                  | VARIABLE
                  | COMPONENT
                  | LABEL"""
    p[0]=common_rule(p[1:]) #pass

#---- component declaration
def p_decl_component(p):
    """decl_component : COMPONENT ID generic_opt port_opt END COMPONENT
                      | COMPONENT ID generic_opt port_opt END COMPONENT ID
                      | COMPONENT ID IS generic_opt port_opt END COMPONENT
                      | COMPONENT ID IS generic_opt port_opt END COMPONENT ID"""
    p[0]=Component(p, id=p[2])

#---- function declaration
def p_decl_function(p):
    """decl_function : IMPURE FUNCTION ID function_decl_opt RETURN typeid function_body_opt
                     | FUNCTION ID function_decl_opt RETURN typeid function_body_opt"""
    impure=0
    if p[1]=='impure': impure=1
    p[0]=Function(p, id=p[impure+2])

def p_function_decl_opt(p):
    """function_decl_opt : '(' parameter_expr_list ')'
                         | empty"""
    p[0]=common_rule(p[1:]) #pass

def p_function_body_opt(p):
    """function_body_opt : IS BEGIN DOT END FUNCTION ID
                         | IS BEGIN DOT END ID
                         | empty"""
    p[0]=common_rule(p[1:]) #pass

#---- procedure declaration
def p_decl_procedure(p):
    "decl_procedure : PROCEDURE ID function_decl_opt procedure_body_opt"
    p[0]=Procedure(p, id=p[2])

def p_procedure_body_opt(p):
    """procedure_body_opt : IS BEGIN sequential_stmt_list_opt END PROCEDURE ID
                          | IS BEGIN sequential_stmt_list_opt END ID
                          | empty"""
    p[0]=common_rule(p[1:]) #pass

#---- for declaration
def p_decl_for(p):
    "decl_for : FOR decl_for_which ':' ID USE decl_for_ent_conf map_opt ';'"
    p[0]=p[6]

def p_decl_for_which(p):
    """decl_for_which : ID
                      | OTHERS
                      | ALL"""
    p[0]=common_rule(p[1:]) #pass

def p_decl_for_ent_conf(p):
    """decl_for_ent_conf : ENTITY ID DOT ID '(' ID ')'
                         | ENTITY ID DOT ID
                         | ENTITY ID '(' ID ')'
                         | ENTITY ID
                         | CONFIGURATION ID DOT ID
                         | CONFIGURATION ID"""
    if p[1]=='entity':
        p[0]=UseEntity(p)
    else:
        p[0]=UseConfiguration(p)

###############################################################################
# 3. Expression
###############################################################################

#-------- expression
def p_expr(p):
    """expr : relexpr
            | expr NAND relexpr
            | expr OR relexpr
            | expr NOR relexpr
            | expr XOR relexpr
            | expr XNOR relexpr
            """
    if len(p)>2:
        p[0]=OpExpr(p, operator=p[2], operands=[p[1], p[3]])
    else:
        p[0]=p[1]

def p_expr1(p):
    "expr : expr AND relexpr"
    # possibly:
    #   substitute (and, SignalEvent(s), (==, s, 1)) -> SignalPosedge(s)
    #   substitute (and, SignalEvent(s), (==, s, 0)) -> SignalNegedge(s)
    if type(p[1])==Prim and type(p[3])==OpExpr:
        ip=1
        ie=3
    elif type(p[1])==OpExpr and type(p[3])==Prim:
        ip=3
        ie=1
    else:
        p[0]=OpExpr(p, operator='and', operands=[p[1], p[3]])
        return
    if type(p[ip].prim)==SignalEvent:
        if p[ie].operator=='==':
            o1=p[ie].operands[0]
            o2=p[ie].operands[1]
            if type(o1)==Prim and type(o2)==Prim:
                if type(o1.prim)==int and o2.prim==p[ip].prim.signal:
                    if o1.prim:
                        se=SignalPosedge(p, signal=o2.prim)
                    else:
                        se=SignalNegedge(p, signal=o2.prim)
                elif type(o2.prim)==int and o1.prim==p[ip].prim.signal:
                    if o2.prim:
                        se=SignalPosedge(p, signal=o1.prim)
                    else:
                        se=SignalNegedge(p, signal=o1.prim)
                else:
                    p[ip].vhdl_error("unsupported use of signal'event")
                p[0]=Prim(p, prim=se)
            else:
                p[ip].vhdl_error("unsupported use of signal'event")
        else:
            p[ip].vhdl_error("unsupported use of signal'event")
    else:
        p[0]=OpExpr(p, operator='and', operands=[p[1], p[3]])
    

#-------- relexpr
def p_relexpr(p):
    """relexpr : sexpr
               | relexpr EQ sexpr
               | relexpr NEQ sexpr
               | relexpr LT sexpr
               | relexpr LE sexpr
               | relexpr GT sexpr
               | relexpr GE sexpr
               """
    if len(p)>2:
        p[0]=OpExpr(p, operator=p[2], operands=[p[1], p[3]])
    else:
        p[0]=p[1]

#-------- sexpr
def p_sexpr(p):
    """sexpr : mulexpr
             | sexpr '+' mulexpr
             | sexpr '-' mulexpr
             | sexpr '&' mulexpr"""
    if len(p)>2:
        p[0]=OpExpr(p, operator=p[2], operands=[p[1], p[3]])
    else:
        p[0]=p[1]

#-------- mulexpr
def p_mulexpr(p):
    """mulexpr : expexpr
               | mulexpr '*' expexpr
               | mulexpr '/' expexpr
               | mulexpr MOD expexpr
               | mulexpr REM expexpr"""
    if len(p)>2:
        p[0]=OpExpr(p, operator=p[2], operands=[p[1], p[3]])
    else:
        p[0]=p[1]

#-------- expexpr
def p_expexpr(p):
    """expexpr : factor
               | expexpr EXPSIGN factor"""
    if len(p)>2:
        p[0]=OpExpr(p, operator=p[2], operands=[p[1], p[3]])
    else:
        p[0]=p[1]

#-------- factor
def p_factor1(p):
    "factor : '-' prim"
    p[0]=OpExpr(p, operator='neg', operands=[p[2]])
def p_factor2(p):
    "factor : '+' prim"
    p[0]=p[2]
def p_factor3(p):
    "factor : prim"
    p[0]=p[1]
def p_factor4(p):
    """factor : NOT prim
              | ABS prim"""
    p[0]=OpExpr(p, operator=p[1], operands=[p[2]])

#-------- prim
def p_prim(p):
    """prim : INTLITERAL
            | LITERAL
            | '(' choice_expr_list ')'
            | NEW ID
            | NEW ID APOSTROPHE '(' expr ')'
            | time
            | expr_object
            """
    if p[1]=='new': # NEW ID ...
        p[0]=c=Prim(p)
        c.ignore(common_rule(p[1:]), "'new' unsupported")
    elif type(p[1])==Time: # time
        p[0]=c=Prim(p)
        c.ignore(p[1], 'time')
    elif p[1]=='(': # '(' choice_expr_list ')'
        p[0]=p[2]
    elif type(p[1]) in [str,int]: # INTLITERAL, LITERAL
        p[0]=c=Prim(p, prim=int(p[1]), str=p[1])
    else: # expr_object
        p[0]=p[1]

def p_expr_object(p):
    """expr_object : id_object
                   | expr_rec_item"""
    p[0]=p[1]

def p_id_object1(p):
    "id_object : ID"
    p[0]=Prim(p, prim=search_id(p, p[1], [Signal, Constant, Parameter]),
            str=p[1])

def p_id_object2(p):
    "id_object : ID APOSTROPHE ID"
    s=search_id(p, p[1], [Signal])
    if s and p[3]=='event':
        p[0]=Prim(p, prim=SignalEvent(p, signal=s), \
            str="%s'event" % p[1])
    else:
        p[0]=Prim(p, str="%s'%s" % (p[1],p[3]))
        p[0].vhdl_error('unsupported event %s' % p[0].str)

def p_id_object3(p):
    "id_object : ID APOSTROPHE '(' expr ')'"
    p[0]=c=Prim(p, str="%s'(something)" % p[1])
    c.ignore(common_rule(p[1:]), \
        "construction unsupported with identificator %s" % p[1])

def p_id_object__range(p):
    "id_object : ID '(' range ')'"
    p[0]=c=Prim(p, str="%s(something)" % p[1])
    prim=search_id(p, p[1], [IdObject])
    if Signal in type(prim).mro():
        c.vhdl_error('indexing is unsupported')
    elif Constant in type(prim).mro():
        c.vhdl_error('indexing of constant is unsupported')
    else:
        # TODO: may be one day, support for conv_std_logic_vector
        c.vhdl_error('unsupported function call')

def p_id_object__parid_list(p):
    "id_object : ID '(' parid_list ')'"
    p_id_object__range(p)

def p_expr_rec_item(p):
    """expr_rec_item : expr_rec_item DOT id_object
                     | id_object DOT id_object"""
    p[0]=c=VHDLobject(p)
    c.ignore(common_rule(p[1:]), 'construction unsupported')

#---- choice to expression list
def p_choice_expr_list1(p):
    "choice_expr_list : choice_list CONNECT expr"
    if p[1]=='others':
        p[0]=p[3]
    else:
        p[0]=c=VHDLobject(p)
        c.ignore(common_rule(p[1:]), 'construction unsupported')

def p_choice_expr_list2(p):
    "choice_expr_list : expr"
    p[0]=p[1]

def p_choice_expr_list3(p):
    "choice_expr_list : choice_list CONNECT expr ',' choice_expr_list"
    p[0]=c=VHDLobject(p)
    c.ignore(common_rule(p[1:]), 'construction unsupported')

def p_choice_expr_list4(p):
    "choice_expr_list : expr ',' choice_expr_list"
    p[0]=c=VHDLobject(p)
    c.ignore(common_rule(p[1:]), 'construction unsupported')

#---- key parameter list (or index)
def p_parid_list(p):
    """parid_list : parid ',' parid_list
                  | parid"""
    p[0]=common_rule(p[1:]) #pass

#---- key parameter (or index)
def p_parid(p):
    """parid : expr
             | ID CONNECT expr"""
    p[0]=common_rule(p[1:]) #pass

###############################################################################
# 4. Sequential Statements
###############################################################################

# match: [postponed] (VHDL93)
def p_postponed_opt(p):
    """postponed_opt : POSTPONED
                     | empty"""
    p[0]=common_rule(p[1:]) #pass

#-------- sequential statement list

# returns: ([CondAssign*], {Signal:[Signal*]} (i.e., sensitivity list))

def p_sequential_stmt_list_opt(p):
    """sequential_stmt_list_opt : sequential_stmt_list
                                | empty"""
    p[0]=p[1]

def p_sequential_stmt_list1(p):
    "sequential_stmt_list : sequential_stmt ';'"
    p[0]=p[1]
def p_sequential_stmt_list2(p):
    "sequential_stmt_list : sequential_stmt ';' sequential_stmt_list"
    p[0]=p[1]+p[3]

def p_sequential_stmt(p):
    """sequential_stmt : seq_stmt_wait
                       | seq_stmt_assert
                       | seq_stmt_report
                       | seq_stmt_assign
                       | seq_stmt_varassign
                       | seq_stmt_proc_call
                       | seq_stmt_if
                       | seq_stmt_case
                       | seq_stmt_while
                       | seq_stmt_for
                       | seq_stmt_next
                       | seq_stmt_exit
                       | seq_stmt_return
                       | seq_stmt_null"""
    if type(p[1])==Assign: # derived from seq_stmt_assign?
        p[0]=[CondAssign(p, guard=Prim(p=None, prim=1), action=p[1])]
    elif type(p[1])==list:
        # CondAssign derived from seq_stmt_if (seq_stmt_case?)
        # every other type of statement is ignored
        p[0]=[i for i in p[1] if type(i)==CondAssign]
    elif type(p[1])==SeqProcCall:
        p[1].ignore(p[1].name, 'procedure call')
        p[0]=[]
    else:
        p[0]=[]

#---- seq statement wait
def p_seq_stmt_wait(p):
    """seq_stmt_wait : WAIT ON id_list UNTIL expr FOR expr
                     | WAIT ON id_list UNTIL expr
                     | WAIT ON id_list FOR expr
                     | WAIT ON id_list
                     | WAIT UNTIL expr FOR expr
                     | WAIT UNTIL expr
                     | WAIT FOR expr
                     | WAIT"""
    p[0]=c=Wait(p)
    if len(p)>2:
        c.ignore(common_rule(p[2:]), 'wait_statement')

#---- seq statement assert
def p_seq_stmt_assert(p):
    """seq_stmt_assert : ASSERT expr REPORT LITERAL severity
                       | ASSERT expr REPORT LITERAL
                       | ASSERT expr severity
                       | ASSERT expr"""
    p[0]=c=Assert(p)
    c.ignore(common_rule(p[2:]), 'assert_expr')

def p_severity(p):
    """severity : SEVERITY NOTE
                | SEVERITY WARNING
                | SEVERITY ERROR
                | SEVERITY FAILURE"""
    p[0]=common_rule(p[1:]) #pass

#---- seq statement report (VHDL93)
def p_seq_stmt_report(p):
    """seq_stmt_report : REPORT LITERAL severity
                       | REPORT LITERAL"""
    p[0]=c=Report(p)
    c.ignore(common_rule(p[2:]), 'report_expr')

#---- seq statement signal assign
def p_seq_stmt_assign(p):
    """seq_stmt_assign : seq_stmt_proc_call_form LE TRANSPORT expr_list
                       | seq_stmt_proc_call_form LE INERTIAL expr_list
                       | seq_stmt_proc_call_form LE expr"""
    p[0]=c=Assign(p, parent=get_parent())
    if p[3] in ['transport', 'inertial']:
        c.vhdl_error("'transport' or 'inertial' unsupported")
    elif Expr in type(p[3]).mro():
        signame=p[1][0]
        if type(signame)==list:
            c.vhdl_error('indexing of signal unsupported')
        else:
            c.signal=p[1][0]
            c.signal.usage.append(Usage(p, access=Usage.W))
            c.value=p[3]
            for s in c.value.get_evals():
                s.usage.append(Usage(p, access=Usage.R))
    else:
        c.vhdl_error('unsupported assignment')

#---- seq statement variable assign
def p_seq_stmt_varassign(p):
    "seq_stmt_varassign : ID ASSIGN expr"
    p[0]=c=VarAssign(p)
    c.ignore(common_rule(p[1:]), 'variable assignment')

#---- seq statement proc call
def p_seq_stmt_proc_call(p):
    "seq_stmt_proc_call : seq_stmt_proc_call_form"
    p[0]=c=SeqProcCall(p, name=p[1][0])
    c.ignore(p[1], 'procedure call')

def p_seq_stmt_proc_call_form(p):
    """seq_stmt_proc_call_form : ID '(' id_map_key_opt_list ')'
                               | ID"""
    p[0]=p[1:]

#---- seq statement if
def p_seq_stmt_if(p):
    """seq_stmt_if : ID ':' IF expr THEN sequential_stmt_list seq_stmt_elsif_list_opt seq_stmt_else_opt END IF ID
                   | ID ':' IF expr THEN sequential_stmt_list seq_stmt_elsif_list_opt seq_stmt_else_opt END IF
                   | IF expr THEN sequential_stmt_list seq_stmt_elsif_list_opt seq_stmt_else_opt END IF ID
                   | IF expr THEN sequential_stmt_list seq_stmt_elsif_list_opt seq_stmt_else_opt END IF"""
    # gets: guard1, [CondAssign*], [CondAssign*],n_guard   [CondAssign*]
    i=0
    if p[2]==':': i=2
    guard1 = p[2+i]
    sl1=list(guard1.get_evals())
    for s in guard1.get_evals():
        s.usage.append(Usage(p, access=Usage.R))
    stmt1 = p[4+i]
    stmt2,guard2 = p[5+i]
    sl2=list(guard2.get_evals())
    stmt3 = p[6+i]
    # if else statement is empty...
#    if not stmt3:
#        a=Assign(p, parent=get_parent())
#        a.signal=stmt1[0].action.signal
#        a.value=Prim(p=None, prim=a.signal)
#        stmt3=[CondAssign(p, guard=Prim(p=None, prim=1), action=a)]
    # gives: [CondAssign*]
    for i in range(len(stmt1)):
        stmt1[i].guard = and_expr([guard1, stmt1[i].guard])
    guard1 = not_expr(guard1)
    for i in range(len(stmt2)):
        stmt2[i].guard = and_expr([guard1, stmt2[i].guard])
    guard1 = and_expr([guard1, guard2])
    for i in range(len(stmt3)):
        stmt3[i].guard = and_expr([guard1, stmt3[i].guard])
    # update sens_list
    for s in stmt1:
        s.sens_list = sl1+s.sens_list
    for s in stmt2:
        s.sens_list = sl1+s.sens_list
    for s in stmt3:
        s.sens_list = sl1+sl2+s.sens_list
    p[0]=stmt1+stmt2+stmt3

def p_seq_stmt_elsif_list_opt(p):
    """seq_stmt_elsif_list_opt : seq_stmt_elsif_list
                               | empty"""
    if p[1]:
        p[0]=p[1]
    else:
        p[0]=([], Prim(p=None, prim=1))

def p_seq_stmt_elsif_list1(p):
    "seq_stmt_elsif_list : seq_stmt_elsif"
    # gets: ([CondAssign*],not(guard))
    p[0]=p[1]
def p_seq_stmt_elsif_list2(p):
    "seq_stmt_elsif_list : seq_stmt_elsif seq_stmt_elsif_list"
#if c1 then
#  [cc1,a1]*
#elsif c2 then \                                     \
#  [cc2,a2]*   /([c2&cc2,a2],!c2)                    |
#elsif c3 then \                  \                  |([c2&cc2,a2],
#  [cc3,a3]*   /([c3&cc3,a3],!c3) /([c3&cc3,a3],!c3) / [!c2&c3&cc3,a3];!c2&!c3)
#else
#  [cc4,a4]*
    # gets: ([CondAssign*],not(guard)) ([CondAssign*],not(guard))
    stmt1,guard1 = p[1]
    stmt2,guard2 = p[2]
    stmt=stmt1
    for i in range(len(stmt2)):
        stmt2[i].guard = and_expr([guard1, stmt2[i].guard])
    stmt+=stmt2
    sl=list(guard1.get_evals())
    for s in stmt2:
        s.sens_list = sl+s.sens_list
    p[0]=(stmt, and_expr([guard1, guard2]))

def p_seq_stmt_elsif(p):
    "seq_stmt_elsif : ELSIF expr THEN sequential_stmt_list"
    # change seq_stmt_list [CondAssign(g,a)*] such that new CondAssign will be:
    # CondAssign(expr and g, a) and return
    # output: ([changed_CondAssign*],not(expr))

    # update usage
    guard=p[2]
    for s in guard.get_evals():
        s.usage.append(Usage(p, access=Usage.R))
    # condition actions
    stmt=p[4]
    sl=list(guard.get_evals())
    for i in range(len(stmt)):
        stmt[i].guard=and_expr([guard, stmt[i].guard])
        stmt[i].sens_list = sl+stmt[i].sens_list
    p[0]=(stmt, not_expr(guard))

def p_seq_stmt_else_opt(p):
    """seq_stmt_else_opt : ELSE sequential_stmt_list
                         | empty"""
    if p[1]:
        p[0]=p[2]
    else:
        p[0]=[]

#---- seq statement case
def p_seq_stmt_case(p):
    """seq_stmt_case : ID ':' CASE expr IS seq_stmt_case_when_list END CASE ID
                     | ID ':' CASE expr IS seq_stmt_case_when_list END CASE
                     | CASE expr IS seq_stmt_case_when_list END CASE"""
    p[0]=c=VHDLobject(p)
    c.vhdl_error("'case' is unsupported construction")

def p_seq_stmt_case_when_list(p):
    """seq_stmt_case_when_list : WHEN choice_list CONNECT
                               | WHEN choice_list CONNECT sequential_stmt_list"""
    p[0]=common_rule(p[1:]) #pass

#---- seq statement while
def p_seq_stmt_while(p):
    """seq_stmt_while : ID ':' WHILE expr LOOP sequential_stmt_list END LOOP ID
                      | ID ':' LOOP sequential_stmt_list END LOOP ID
                      | WHILE expr LOOP sequential_stmt_list END LOOP
                      | LOOP sequential_stmt_list END LOOP"""
    p[0]=c=VHDLobject(p)
    c.vhdl_error("'loop' is unsupported construction")

#---- seq statement for
def p_seq_stmt_for(p):
    """seq_stmt_for : ID ':' FOR ID IN range LOOP sequential_stmt_list END LOOP ID
                    | ID ':' FOR ID IN range LOOP sequential_stmt_list END LOOP
                    | FOR ID IN range LOOP sequential_stmt_list END LOOP"""
    p[0]=c=VHDLobject(p)
    c.vhdl_error("'loop' is unsupported construction")

#---- seq statement next
def p_seq_stmt_next(p):
    """seq_stmt_next : NEXT ID WHEN expr
                     | NEXT ID
                     | NEXT WHEN expr
                     | NEXT"""
    p[0]=c=VHDLobject(p)
    c.ignore(common_rule(p[1:]))

#---- seq statement exit
def p_seq_stmt_exit(p):
    """seq_stmt_exit : EXIT ID WHEN expr
                     | EXIT ID
                     | EXIT WHEN expr
                     | EXIT"""
    p[0]=c=VHDLobject(p)
    c.ignore(common_rule(p[1:]))

#---- seq statement return
def p_seq_stmt_return(p):
    """seq_stmt_return : RETURN expr
                       | RETURN"""
    p[0]=c=VHDLobject(p)
    c.ignore(common_rule(p[1:]))

#---- seq statement null
def p_seq_stmt_null(p):
    "seq_stmt_null : NULL"
    p[0]=VHDLobject(p)

###############################################################################
# 5. Parallel Statements
###############################################################################

#-------- parallel statement list
def p_parallel_stmt_list_opt(p):
    """parallel_stmt_list_opt : parallel_stmt_list
                              | empty"""
    if p[1]:
        p[0]=p[1]
    else:
        p[0]=[]

def p_parallel_stmt_list(p):
    """parallel_stmt_list : parallel_stmt ';'
                          | parallel_stmt ';' parallel_stmt_list"""
    if type(p[1])==list: # can be list of parallel_stmts (see par_stmt_if)
        if len(p)>3:
            p[0]=p[1]+p[3]
        else:
            p[0]=p[1]
    else:
        generate_list(p, 1, 3)

#-------- parallel statement
def p_parallel_stmt(p):
    """parallel_stmt : par_stmt_block
                     | par_stmt_process
                     | par_stmt_proc_call
                     | par_stmt_assign
                     | par_stmt_assert
                     | par_stmt_with
                     | par_stmt_comp
                     | par_stmt_entity
                     | par_stmt_configuration
                     | par_stmt_if
                     | par_stmt_for"""
    p[0]=p[1]

#---- parallel statement block
def p_par_stmt_block(p):
    "par_stmt_block : ID ':' BLOCK is_opt par_generic_opt par_port_opt declaration_list BEGIN parallel_stmt_list_opt END BLOCK id_opt"
    p[0]=c=VHDLobject(p)
    c.ignore(common_rule(p[1:]))

def p_par_generic_opt(p):
    """par_generic_opt : GENERIC '(' id_type_list ')' ';' generic_map_key_opt
                       | empty"""
    p[0]=common_rule(p[1:]) #pass
                       
def p_par_port_opt(p):
    """par_port_opt : PORT '(' id_port_type_list ')' ';' port_map_key_opt
                    | empty"""
    p[0]=common_rule(p[1:]) #pass
                       
#---- parallel statement process
def p_par_stmt_process1(p):
    "par_stmt_process : postponed_opt PROCESS process_sens_list_opt declaration_list BEGIN sequential_stmt_list_opt END postponed_opt PROCESS id_opt"
    p[0]=c=Process(p, sens_list=p[3], assigns=p[6])
def p_par_stmt_process2(p):
    "par_stmt_process : ID ':' postponed_opt PROCESS process_sens_list_opt declaration_list BEGIN sequential_stmt_list_opt END postponed_opt PROCESS id_opt"
    p[0]=c=Process(p, sens_list=p[5], assigns=p[8])

def p_process_sens_list_opt(p):
    """process_sens_list_opt : '(' id_list ')'
                             | empty"""
    if p[1]:
        p[0]=p[2]
    else:
        p[0]=[]

#---- parallel statement proc call
def p_par_stmt_proc_call1(p):
    "par_stmt_proc_call : ID ':' par_stmt_proc_call_body"
    p[0]=p[3]
def p_par_stmt_proc_call2(p):
    "par_stmt_proc_call : par_stmt_proc_call_body"
    p[0]=p[1]
def p_par_stmt_proc_call3(p):
    "par_stmt_proc_call : POSTPONED par_stmt_proc_call_body"
    p[0]=p[2]
def p_par_stmt_proc_call4(p):
    "par_stmt_proc_call : ID ':' POSTPONED par_stmt_proc_call_body"
    p[0]=p[4]

def p_par_stmt_proc_call_body(p):
    "par_stmt_proc_call_body : par_stmt_proc_call_form"
    p[0]=c=ProcCall(p)
    c.ignore(common_rule(p[1]))

def p_par_stmt_proc_call_form(p):
    """par_stmt_proc_call_form : ID '(' id_map_key_opt_list ')'
                               | ID '(' range ')'"""
    p[0]=p[1:]

#---- parallel statement signal assign
# VHDL93 version
def p_par_stmt_assign1(p):
    "par_stmt_assign : ID ':' POSTPONED par_stmt_assign_part"
    p[0]=p[4]
def p_par_stmt_assign2(p):
    "par_stmt_assign : ID ':' par_stmt_assign_part"
    p[0]=p[3]
def p_par_stmt_assign3(p):
    "par_stmt_assign : POSTPONED par_stmt_assign_part"
    p[0]=p[2]
def p_par_stmt_assign4(p):
    "par_stmt_assign : par_stmt_assign_part"
    p[0]=p[1]

def p_par_stmt_assign_part1(p):
    """par_stmt_assign_part : par_stmt_proc_call_form LE TRANSPORT par_stmt_assign_expr
                            | par_stmt_proc_call_form LE INERTIAL par_stmt_assign_expr
                            | par_stmt_proc_call_form LE par_stmt_assign_expr"""
    id=search_id(p, p[1][0], [Signal])
    if id:
        p[0]=id
        id.vhdl_error('indexing of signal unsupported')
    else:
        p[0]=c=VHDLobject(p)
        c.vhdl_error('unsupported assignment')

def p_par_stmt_assign_part2a(p):
    "par_stmt_assign_part : ID LE TRANSPORT par_stmt_assign_expr"
    a=Assign(p, parent=get_parent())
    a.vhdl_error("'transport' unsupported")
    p[0]=c=CondAssign(p, guard=Prim(p=None, prim=1), action=a)
def p_par_stmt_assign_part2b(p):
    "par_stmt_assign_part : ID LE INERTIAL par_stmt_assign_expr"
    a=Assign(p, parent=get_parent())
    a.vhdl_error("'inertial' unsupported")
    p[0]=c=CondAssign(p, guard=Prim(p=None, prim=1), action=a)
def p_par_stmt_assign_part2c(p):
    "par_stmt_assign_part : ID LE par_stmt_assign_expr"
    a=Assign(p, parent=get_parent())
    a.signal=p[1]
    a.signal.usage.append(Usage(p, access=Usage.W))
    a.value=p[3]
    for s in a.value.get_evals():
        s.usage.append(Usage(p, access=Usage.R))
    p[0]=CondAssign(p, guard=Prim(p=None, prim=1), action=a)

def p_par_stmt_assign_expr1(p):
    "par_stmt_assign_expr : signal_value"
    p[0]=p[1]
    if type(p[1])==VHDLobject: # UNAFFECTED
        p[1].vhdl_error("'unaffected' unsupported")
def p_par_stmt_assign_expr2(p):
    "par_stmt_assign_expr : signal_value WHEN expr ELSE par_stmt_assign_expr"
    p[0]=c=VHDLobject(p)
    c.vhdl_error('when-else construction unsupported')

def p_signal_value(p):
    """signal_value : expr_list
                    | UNAFFECTED"""
    if p[1]=='unaffected':
        p[0]=c=VHDLobject(p)
        c.ignore('unaffected')
    else:
        p[0]=p[1]

#---- parallel statement assert
def p_par_stmt_assert(p):
    """par_stmt_assert : ID ':' POSTPONED ASSERT expr report_opt severity_opt
                       | ID ':' ASSERT expr report_opt severity_opt
                       | POSTPONED ASSERT expr report_opt severity_opt
                       | ASSERT expr report_opt severity_opt"""
    p[0]=c=Assert(p)
    c.ignore(common_rule(p[1:]))

def p_report_opt(p):
    """report_opt : REPORT LITERAL
                  | empty"""
    p[0]=common_rule(p[1:]) #pass

def p_severity_opt(p):
    """severity_opt : SEVERITY NOTE
                    | SEVERITY WARNING
                    | SEVERITY ERROR
                    | SEVERITY FAILURE
                    | empty"""
    p[0]=common_rule(p[1:]) #pass

#---- parallel statement with
def p_par_stmt_with(p):
    """par_stmt_with : ID ':' POSTPONED WITH expr SELECT par_stmt_assign_part
                     | ID ':' WITH expr SELECT par_stmt_assign_part
                     | POSTPONED WITH expr SELECT par_stmt_assign_part
                     | WITH expr SELECT par_stmt_assign_part"""
    p[0]=c=VHDLobject(p)
    c.ignore(common_rule(p[1:]))

#---- parallel statement component
def p_par_stmt_comp(p):
    "par_stmt_comp : ID ':' ID map_opt"
    p[0]=c=CompInstance(p)
    c.ignore(common_rule(p[1:]))

#---- parallel statement entity (VHDL93)
def p_par_stmt_entity(p):
    """par_stmt_entity : ID ':' ENTITY ID DOT ID '(' ID ')' map_opt
                       | ID ':' ENTITY ID DOT ID map_opt
                       | ID ':' ENTITY ID '(' ID ')' map_opt
                       | ID ':' ENTITY ID map_opt"""
    p[0]=c=EntityImpl(p)
    c.ignore(common_rule(p[1:]))

#---- parallel statement configuraton (VHDL93)
def p_par_stmt_configuration(p):
    """par_stmt_configuration : ID ':' CONFIGURATION ID DOT ID map_opt
                              | ID ':' CONFIGURATION ID map_opt"""
    p[0]=c=VHDLobject(p)
    c.ignore(common_rule(p[1:]))

#---- parallel statement if
def p_par_stmt_if(p):
    """par_stmt_if : ID ':' IF expr GENERATE parallel_stmt_list_opt END GENERATE ID
                   | ID ':' IF expr GENERATE parallel_stmt_list_opt END GENERATE"""
    if Expr not in type(p[4].eval()).mro():
        p[0]=p[6]
    else:
        p[0]=c=CondGenerate(p)
        c.ignore(common_rule(p[1:]))
    
#---- parallel statement for
def p_par_stmt_for(p):
    """par_stmt_for : ID ':' FOR ID IN range GENERATE parallel_stmt_list_opt END GENERATE ID
                    | ID ':' FOR ID IN range GENERATE parallel_stmt_list_opt END GENERATE"""
    p[0]=c=ForGenerate(p)
    c.ignore(common_rule(p[1:]))

###############################################################################
# 6. Others
###############################################################################
def p_error(p):
    if p:
        VHDLobject(p).vhdl_error("invalid syntax `%s'" % (p.value))
#        print "%s:%d:invalid syntax `%s'" % \
#            (p.lexer.filename, p.lexer.lineno, p.value)
    else:
        print '%s:%d:unexpected EOF' % (lexer.filename, lexer.lineno)
    yacc.errok()

# build the parser
import yacc
yacc.yacc(method='LALR')

entities=[]

###############################################################################
# READER
###############################################################################

# switch comments between: abstract begin ... end abstract
def uncomment_abstract_specification(content):
    lines=content.split('\n')
    abstract_spec=False
    for i in range(len(lines)):
        if not abstract_spec:
            if lines[i].strip().startswith('-- abstract begin'):
                abstract_spec=True
        else:
            if lines[i].strip().startswith('-- abstract end'):
                abstract_spec=False
            else:
                if lines[i].strip().startswith('--'):
                    index=lines[i].find('--')
                    lines[i]=lines[i][:index]+lines[i][index+2:]
                else:
                    lines[i]='-- %s' % lines[i]
    return '\n'.join(lines)

def parse_file(filename, abstract_content=None):
    """parse_file(filename, abstract_content=None) parse file. Use abstract
    content (if available) instead of file content."""
    import os
    lexer.filename=filename
    if not abstract_content:
        if os.access(filename, os.R_OK):
            f=open(filename)
            abstract_content=f.read()
            f.close()
    yacc.parse(abstract_content)

def parse_abstract_file(filename):
    try:
        f=open(filename)
        content=f.read()
        f.close()
        content=uncomment_abstract_specification(content)
        return parse_file(filename, content)
    except IOError:
        raise IOError('could not read %s' % filename)

if __name__ == "__main__":
    import sys
    if len(sys.argv)>1:
        parse_file_abstract(sys.argv[1])
    else:
        print 'syntax: %s file.vhd' % sys.argv[0]
