#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:expandtab:sw=4:tw=0
###############################################################################
# vhdltrans.py - transformation from VHDL structures to structures for        #
#                counter automaton generation                                 #
#                                                                             #
# $Id: vhdltrans.py 202 2007-06-10 01:11:52Z smrcka $
#                                                                             #
# Author: Ales Smrcka <smrcka@fit.vutbr.cz>                                   #
#                                                                             #
# This program is free software; you can redistribute it and/or modify        #
#   it under the terms of the GNU General Public License as published by      #
#   the Free Software Foundation; either version 2 of the License, or         #
#   (at your option) any later version.                                       #
#   See http://www.gnu.org/licenses/gpl.html for more details.                #
#                                                                             #
###############################################################################

import vhdl_parser as parser
from simplify import *

class TransformationError(Exception):
    pass

# returns simplify.Variable object according to parser.IdObject,
# creates new simplify.Variable if no reference is in lookup table
# (i.e. dictionary parser.IdObject->simplify.Variable)
# input: parser.IdObject, lookup table
# output: simplify.VariableUse (and created simplify.Variable in lookup_table)
def _vhdl2simplify_id(key, lookup_table):
    # handle: Signal, Parameter, Port
    def signal_ref(key):
        if not ((parser.Signal in type(key).mro()) or \
               (parser.Parameter in type(key).mro())):
            if parser.VHDLobject in type(key).mro():
                key.vhdl_error('unknown object')
            else:
                print key
                print type(key)
                assert (parser.Signal in type(key).mro()) or \
                       (parser.Parameter in type(key).mro())
        if lookup_table.has_key(key):
            v=lookup_table[key]
            return VariableUse(v, 'normal')
        else:
            v=Variable()
            v.vhdl=key
            v.flag='unknown'
            v.name=key.id
            v.comment=None
            try: sens_list=key.sens_list
            except: sens_list=[]
            t=key.get_type()
            if t==parser.Int:
                v.type='int'
            elif t==parser.Bool:
                v.type='bool'
            elif type(t)==parser.BinVectorType:
                # BinVectors are known as int
                v.type='int'
            else: # parser.CommonType
                # CommonType is supported only for parameters as int
                if type(key)==parser.Parameter:
                    v.type='int'
                else:
                    key.vhdl_error('unknown type %s' % t)
            lookup_table[v.vhdl]=v # enlist new variable
            # parser.Parameter may have extra expression
            if parser.Parameter in type(key).mro():
                if key.constraint:
                    v.comment = _vhdl2simplify_expression( \
                                    key.constraint.get_list(), lookup_table)
            return VariableUse(v, 'normal')
    # handle: SignalPosedge, SignalNegedge
    def signalproperty_ref(property):
        assert type(property.signal)!=parser.SignalPosedge
        v=signal_ref(property.signal).v
        v.flag='state clock'
        if type(property)==parser.SignalPosedge:
            return VariableUse(v, 'posedge')
        else:
            return VariableUse(v, 'negedge')

    if type(key)==parser.SignalEvent:
        key.trans_error("%s'event unsupported " % key.signal.id)
    elif type(key) in [parser.SignalPosedge,parser.SignalNegedge]:
        return signalproperty_ref(key)
    else: # Signal, Port, Parameter
        return signal_ref(key)

# input: VHDL Expr.get_list object, lookup table
# output: simplify expression object with substituted prims
def _vhdl2simplify_expression(lexpr, lookup_table):
    if type(lexpr)==int:
        return lexpr
    elif type(lexpr)==list:
        for i in xrange(1,len(lexpr)):
            lexpr[i] = _vhdl2simplify_expression(lexpr[i], lookup_table)
        return lexpr
    else: # IdObject
        return _vhdl2simplify_id(lexpr, lookup_table)

# Behavior object representing a tripple: guard -> dest:=value
#   guard - expression over Variable objects (for simplify module)
#   var   - Variable object reference
#   value - expression over Variable objects (for simplify module)
# plus other properties:
#   sensitive - sensitivity list of variable's name
#   vhdl_source - (filename,lineno,str_line)
class Behavior(object):
    nonvar=Variable()
    nonvar.flag='none'
    def __init__(self, guard, var, value):
        self.guard=guard
        if var==None:
            self.var=self.nonvar
        else:
            self.var=var
        self.value=value
        self.sensitive=[]
        self.vhdl_source=('',0,'')

    def __eq__(self, other):
        return str(self)==str(other)

    def __str__(self):
        if self.var==self.nonvar:
            return '%s -> true' % strexpr(self.guard)
        else:
            return '%s -> %s:=%s' % \
                (strexpr(self.guard), self.var, strexpr(self.value))

    def copy(self):
        b=Behavior(copy_expression(self.guard), self.var, \
                     copy_expression(self.value))
        b.sensitive=[i for i in self.sensitive]
        b.vhdl_source=self.vhdl_source
        return b

# From VHDL structures, read all CondAssigns and transform them to Behavior
# objects, also generate set of Variables
# input: VHDL entity
# output: (set of Variables, set of Behavior objects)
#   Variables.flag set to 'unknown'
#   VariableUse.use ∈ {'normal', 'posedge', 'negedge'}
#   Behavior.guard - expression over VariableUse or int
#   Behavior.var - Variable reference
#   Behavior.value - expression over variableUse or int
def read_useful_behavior(entity):
    lookup_table={}
    behavior=set()
    for i in entity.architecture.connections:
        guard=_vhdl2simplify_expression(i.guard.get_list(), lookup_table)
        guard=evaluate(guard, {})
        var=_vhdl2simplify_id(i.action.signal, lookup_table).v
        value=_vhdl2simplify_expression(i.action.value.get_list(), lookup_table)
        b=Behavior(guard, var, value)
        sensitive=[_vhdl2simplify_id(s, lookup_table) for s in i.sens_list]
        for s in sensitive:
            if s not in b.sensitive: b.sensitive.append(s)
        if i.filename:
            f=open(i.filename)
            line=f.readlines()[i.lineno-1].strip()
            f.close()
            b.vhdl_source=(i.filename,i.lineno,line)
        behavior.add(b)
    return (set(lookup_table.values()), behavior)
