#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:expandtab:sw=4:tw=0
###############################################################################
# output_blast.py - output module for BLAST                                   #
#                                                                             #
# $Id: output_armc.py 201 2007-06-10 00:54:39Z smrcka $
#                                                                             #
# Author: Ales Smrcka <smrcka@fit.vutbr.cz>                                   #
#                                                                             #
# This program is free software; you can redistribute it and/or modify        #
#   it under the terms of the GNU General Public License as published by      #
#   the Free Software Foundation; either version 2 of the License, or         #
#   (at your option) any later version.                                       #
#   See http://www.gnu.org/licenses/gpl.html for more details.                #
#                                                                             #
###############################################################################

def get_state_map(states):
    "get_state_map(states) returns map State->int"
    state_map = {}
    i = 0
    for state in states:
        state_map[state] = i
        i += 1
    return state_map

# operator map (simplify->C)
opmap = {
    '-':    '-',
    '+':    '+',
    '==':   '==',
    '!=':   '!=',
    '<':    '<',
    '>':    '>',
    '<=':   '<=',
    '>=':   '>=',
    '*':    '*',
    'and':  '&&',
    'or':   '||',
    'not':  '!',
}

def expr2str(expr):
    "expr2str(expr) returns string representation of expression in C syntax"
    if type(expr)==list:
        return '(%s)' % opmap[expr[0]].join([expr2str(e) for e in expr[1:]])
    else:
        return str(expr)

def actions2strlist(actions):
    "returns list of pairs (variable name, expression in C)"
    return [(a.counter.name, expr2str(a.value)) for a in actions]


def new_name(var):
    return '__NEW_%s' % var

def transformate_automaton(ca):
    "transformate_automaton(ca) returns automaton as string"
    from time import strftime
    from sys import argv

    # map of State->int (init state first)
    state_map = get_state_map(ca.states)

    # variables declaration
    vars = [v.name for v in list(ca.counters | ca.parameters)]
    variables = vars[:]
    for var in vars:
        variables.append(new_name(var))
    variables = ['int %s;' % v for v in variables]
    variables = '\n    '.join(variables)

    # input predicates
    if ca.output_options.has_key('pred'):
        pred = open(ca.output_options['pred']).read()
    else:
        pred = ''

    # transition relation
    ind = ' '*4
    gtrans = []
    for s in ca.states:
        trans = []
        otransitions = [t for t in ca.transitions if t.s1 == s]
        for t in otransitions:
            action_tuples = actions2strlist(t.actions)
            actions = [action % (new_name(var),value)
                        for var,value in action_tuples]
            actions = actions + [action % (var,new_name(var))
                        for var,value in action_tuples]
            trans.append(transition % {
                'desc': str(t.s2),
                'to_state': state_map[t.s2],
                'guard': expr2str(t.guard),
                'actions': '\n'.join(actions) })

        # infinite loop if no outgoing transition
        if not trans: trans = ['{}']

        gtrans.append(transitions % {
            'desc': str(s),
            'from_state': state_map[s],
            'transitions': trconn.join(trans) })

    return template % {
        'vendor': ca.vendor,
        'model': ca.name,
        'src_files':' '.join(ca.sources),
        'trtime': ca.time,
        'ctime': strftime('%c'),
        'cmdline': ' '.join(argv),
        'variables': variables,
        'pred': pred,
        'init_state': state_map[ca.init_state],
        'bad_state': state_map[ca.bad_state],
        'transition_relation': trsconn.join(gtrans),
    }

template = """/*
This model was automatically generated by %(vendor)s
For more information please contact Ales Smrcka <smrcka@fit.vutbr.cz>
model name: %(model)s
source files: %(src_files)s
transformation time: %(trtime) s
created: %(ctime)s
cmd line: %(cmdline)s
*/

int main()
{
    int state;
    int __BLAST_NONDET;
    /* counter and parameters declaration */
    %(variables)s

    /* init predicates */
    %(pred)s

    /* init state */
    state = %(init_state)d;

    /* FSM */
    while (state != %(bad_state)d)
        %(transition_relation)s
    ERROR:
}
"""

transitions = """
        /* from state %(desc)s */
        if (state == %(from_state)d) {
            while (1) %(transitions)s
        }
"""
trsconn = """
        else
"""

transition = """
                /* to state %(desc)s */
                if (__BLAST_NONDET && %(guard)s) {
                    state = %(to_state)d;
%(actions)s
                    break;
                }
"""
action = '                    %s = %s;'
trconn = """
                else
"""
